# Generated by Django 4.2.7 on 2023-11-28 14:49

from django.db import migrations, models
import django.db.models.deletion
import picklefield.fields


class Migration(migrations.Migration):
    dependencies = [
        ("djwebdapp", "0016_alter_transaction_has_code"),
    ]

    operations = [
        migrations.CreateModel(
            name="Event",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=500)),
                (
                    "args",
                    picklefield.fields.PickledObjectField(
                        blank=True, default=dict, editable=False
                    ),
                ),
                ("event_index", models.PositiveBigIntegerField()),
                ("normalized", models.BooleanField(default=False)),
                (
                    "transaction",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="transactionevent_set",
                        to="djwebdapp.transaction",
                    ),
                ),
            ],
        ),
    ]



from rest_framework import serializers

from fa2.models import Fa2Token
from token_metadata.serializers import Tzip21Serializer


class Fa2TokenSerializer(serializers.ModelSerializer):
    address = serializers.CharField(source='contract.origination.address')
    metadata = Tzip21Serializer()

    class Meta:
        model = Fa2Token
        fields = [
            "address",
            "token_id",
            "metadata",
        ]



# Generated by Django 4.1.1 on 2023-02-09 14:23

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("djwebdapp", "0013_transaction_normalized_dependency"),
    ]

    operations = [
        migrations.AlterUniqueTogether(
            name="account",
            unique_together={("blockchain", "address")},
        ),
    ]



from djwebdapp.models import Account

# Given a blockchain and no address, djwebdapp will create a keypair
admin = Account.objects.create(blockchain=blockchain, name='Admin')

# Now, provision this wallet with some xtz from the default client wallet
client.transaction(
    destination=admin.address,
    amount=1_000_000,
).autofill().sign().inject(_async=False)

# Wait until the account receives the balance
while not admin.balance:
    admin.refresh_balance()



# First, we need to add a blockchain in the database
from djwebdapp.models import Blockchain
blockchain, _ = Blockchain.objects.get_or_create(
    name='Ethereum Local',
    provider_class='djwebdapp_ethereum.provider.EthereumProvider',
    # on geth local blockchain, blocks are mined at every transaction, so we
    # don't need to wait for confirmation blocks
    min_confirmations=0,
)

# Add our node to the blockchain
blockchain.node_set.get_or_create(endpoint='http://ethlocal:8545')



import logging

from eth_utils.abi import event_abi_to_log_topic
from hexbytes import HexBytes
from web3 import Web3
from web3.exceptions import ContractLogicError

from django.conf import settings

from django.db import transaction as db_transaction
from djwebdapp.models import Account
from djwebdapp_ethereum.models import EthereumEvent, EthereumTransaction
from djwebdapp.provider import Provider


class EthereumProvider(Provider):
    logger = logging.getLogger('djwebdapp_ethereum')
    transaction_class = EthereumTransaction

    def generate_secret_key(self):
        wallet = self.client.eth.account.create()
        return wallet.address, bytes(wallet.key)

    def should_activate_client_middleware(self, endpoint):
        """
        For non-mainnet chains, the client needs to activate an onion
        middleware. See:
        https://web3py.readthedocs.io/en/stable/middleware.html#geth-style-proof-of-authority
        This method checks whether, given the blockchain endpoint, this
        middleware should be activated
        """  # noqa: E501
        middleware_endpoints = [
            'ethlocal',
            'localhost',
            'polygon-mumbai',
        ]
        return len([substr in endpoint for substr in middleware_endpoints])

    def get_client(self, **kwargs):
        endpoint = self.blockchain.node_set.first().endpoint
        client = Web3(Web3.HTTPProvider(endpoint))

        if settings.DEBUG and not self.wallet:  # geth default account
            client.eth.default_account = client.eth.accounts[0]

        if self.should_activate_client_middleware(endpoint):
            from web3.middleware import geth_poa_middleware
            client.middleware_onion.inject(geth_poa_middleware, layer=0)

        return client

    def get_balance(self, address=None):
        weis = self.client.eth.get_balance(
            address
            or getattr(self.wallet, 'address', False)
            or self.client.eth.default_account
        )
        return self.client.from_wei(weis, 'ether')

    @property
    def head(self):
        """
        Return the current block number.
        """
        return self.client.eth.get_block_number()

    def index_level(self, level):
        block = self.client.eth.get_block(level, True)
        for transaction in block.transactions:
            to = transaction.get('to', None)
            if to is None and self.check_hash(transaction['hash'].hex()):
                self.index_contract(level, transaction)
            elif to in self.addresses:
                self.index_call(level, transaction)

    def index_contract(self, level, transaction):
        self.logger.info(f'Syncing origination {transaction["hash"]}')
        contract = EthereumTransaction.objects.get(
            blockchain=self.blockchain,
            hash=transaction['hash'].hex(),
        )
        contract.level = level
        contract.gas = transaction['gas']
        contract.metadata = self.json(transaction)
        contract.state_set('done')

    def index_call(self, level, transaction):
        self.logger.info(f'EthereumProvider.index_call({transaction})')

        contract = EthereumTransaction.objects.filter(
            blockchain=self.blockchain,
            address=transaction['to'],
        ).first()

        if not contract:
            contract = EthereumTransaction.objects.create(
                blockchain=self.blockchain,
                address=transaction['to'],
                index=False,
            )

        call = self.transaction_class.objects.select_subclasses().filter(
            hash=transaction['hash'].hex(),
        ).first()

        tx_receipt = self.client.eth.get_transaction_receipt(
            transaction['hash'].hex(),
        )

        if tx_receipt.status == 0:
            return

        if not call:
            call = EthereumTransaction(
                hash=transaction['hash'].hex(),
                contract=contract,
                blockchain=self.blockchain,
            )
        call.metadata = self.json(transaction)
        call.gas = transaction['gas']
        call.level = level
        call.sender, _ = Account.objects.get_or_create(
            address=transaction['from'],
            blockchain=self.blockchain,
        )
        if contract.abi:
            interface = self.client.eth.contract(
                address=contract.address,
                abi=contract.abi,
            )

        if self.is_smart_contract_transfer_only(call):
            call.function = 'receive'
            call.amount = self.client.from_wei(call.metadata['value'], 'ether')
        elif contract.abi:
            fn, args = interface.decode_function_input(call.metadata['input'])
            call.function = fn.fn_name
            call.args = args

        call.state_set('done')

    def is_smart_contract_transfer_only(self, call):
        """
        You can send ETH to smart contract without calling a method by implementing
        a `receive` or `fallback` function. In which case, the input value is
        always '0x' (no data sent to the contract).

        This method returns `True` if the transaction is such a call.

        More information here: https://ethereum.stackexchange.com/questions/130936/contract-write-function-has-no-input-data0x
        """  # noqa: E501
        return call.metadata['input'] == '0x'

    def json(self, transaction):
        return {
            key: value.hex() if isinstance(value, HexBytes) else value
            for key, value in transaction.items()
        }

    def send(self, transaction):
        Contract = self.client.eth.contract(  # noqa
            abi=transaction.contract.abi,
            address=transaction.contract.address,
        )
        funcs = Contract.find_functions_by_name(transaction.function)
        if not funcs:
            raise Exception(
                f'{transaction.function} not found in {transaction.contract}'
            )

        func = funcs[0]

        args = transaction.get_args()

        for i, inp in enumerate(func.abi.get('inputs', [])):
            if inp['type'] == 'tuple':
                tmp_list_from_tuple = []

                for j, component in enumerate(inp['components']):
                    if component['type'].startswith('bytes32'):
                        tmp_list_from_tuple.append(
                            self.client.to_bytes(hexstr=args[i][j])
                        )
                    else:
                        tmp_list_from_tuple.append(args[i][j])
                args[i] = tuple(tmp_list_from_tuple)

            elif inp['type'].startswith('bytes32'):
                args[i] = self.client.to_bytes(hexstr=args[i])

        tx = func(*args)
        transaction.hash = self.write_transaction(transaction, tx)

    def deploy(self, transaction):
        self.logger.debug(f'{transaction}.deploy(): start')
        transaction.level = self.head

        if transaction.kind == 'contract':
            transaction.sender.last_level = self.head
            self.originate(transaction)
            transaction.sender.save()
        elif transaction.kind == 'function':
            transaction.sender.last_level = self.head
            self.send(transaction)
            transaction.sender.save()
        elif transaction.kind == 'transfer':
            transaction.hash = self.transfer(transaction)
        else:
            transaction.error = f'Unknown transaction kind {transaction.kind}'
            transaction.state_set('failed')
            return

        self.logger.info(f'{transaction}.deploy(): success')

    def transfer(self, transaction):
        tx = {
            'to': transaction.receiver.address,
            'from': transaction.sender.address,
            'value': self.client.to_wei(transaction.amount, 'ether'),
        }
        tx['gas'] = self.client.eth.estimate_gas(tx)
        tx['gasPrice'] = self.client.eth.gas_price
        tx['nonce'] = self.client.eth.get_transaction_count(
            transaction.sender.address
        )
        tx['chainId'] = self.client.eth.chain_id
        signed_txn = self.client.eth.account.sign_transaction(
            tx,
            private_key=transaction.sender.get_secret_key(),
        )
        self.client.eth.send_raw_transaction(signed_txn.rawTransaction)
        return self.client.to_hex(
            self.client.keccak(signed_txn.rawTransaction)
        )

    def originate(self, transaction):
        Contract = self.client.eth.contract(  # noqa
            abi=transaction.abi,
            bytecode=transaction.bytecode,
        )

        tx = Contract.constructor(*transaction.get_args())
        transaction.hash = self.write_transaction(transaction, tx)
        receipt = self.client.eth.wait_for_transaction_receipt(
            transaction.hash)
        transaction.address = receipt.contractAddress

    def write_transaction(self, djwebdapp_transaction, tx):
        sender = djwebdapp_transaction.sender
        nonce = self.client.eth.get_transaction_count(sender.address)
        options = {
            'from': sender.address,
            'nonce': nonce,
        }
        try:
            built = tx.build_transaction(options)
        except ContractLogicError as e:
            djwebdapp_transaction.error = e.message
            raise e
        options['gas'] = self.client.eth.estimate_gas(built)
        signed_txn = self.client.eth.account.sign_transaction(
            built,
            private_key=sender.get_secret_key(),
        )

        self.client.eth.send_raw_transaction(signed_txn.rawTransaction)
        return self.client.to_hex(
            self.client.keccak(signed_txn.rawTransaction)
        )


class EthereumEventProvider(EthereumProvider):
    event_class = EthereumEvent

    def download(self, contract_addresses, from_block, to_block):
        self.index(
            contract_addresses,
            from_block,
            to_block,
            save_blockchain=False,  # don't override indexer's state
        )

    def index_init(
        self,
        contract_addresses=None,
        from_block=None,
        to_block=None,
    ):
        super().index_init()
        self.contracts = self.contracts.exclude(
            abi=None,
        )
        if contract_addresses:
            self.contracts = self.contracts.filter(
                address__in=contract_addresses,
            )

        self.addresses = self.contracts.values_list(
            'address',
            flat=True,
        )

        head = self.head

        event_filter = {
            'fromBlock': from_block or self.blockchain.index_level or 0,
            'toBlock': to_block or head,
            'address': self.addresses,
        }

        try:
            logs = self.client.eth.get_logs(event_filter)
        except Exception:
            logs = []
            init_level = self.blockchain.index_level or 0
            number_of_levels_to_increment = self.blockchain.configuration.get(
                'number_of_levels_to_increment',
                50,
            )

            while init_level < head:
                to_block = min(
                    init_level + number_of_levels_to_increment,
                    head,
                )
                event_filter['fromBlock'] = init_level
                event_filter['toBlock'] = to_block
                logs += self.client.eth.get_logs(event_filter)
                init_level = to_block + 1

        self.logs = logs
        self.last_indexed_block = head

    def get_min_level(self, items, key):
        return min((item[key] for item in items), default=None)

    def index(
        self,
        contract_addresses=None,
        from_block=None,
        to_block=None,
        save_blockchain=True,
    ):
        """
        Index EVM blockchains.

        Iterate over each level that included txs and events
        related to indexed contracts.
        """
        self.index_init(contract_addresses, from_block, to_block)

        while len(self.hashes) or len(self.logs):
            index_of_level_in_hashes_tuple = 1
            min_level_hashes = self.get_min_level(
                self.hashes,
                key=index_of_level_in_hashes_tuple,
            )
            min_level_logs = self.get_min_level(
                self.logs,
                key='blockNumber',
            )

            level_to_index = min(
                filter(None, [min_level_hashes, min_level_logs])
            )

            self.logger.info(f'Indexing level {level_to_index}')
            self.index_level(level_to_index)
            self.blockchain.index_level = level_to_index

        self.blockchain.index_level = self.last_indexed_block
        if save_blockchain:
            self.blockchain.save()

    def index_level(self, level):
        """
        To optimize for compute units on Alchemy/Moralis nodes, we:

        #. Filter logs fetched for a range of level at init to restrict them
           to the current level

        #. only query the full block to index relevant transactions if:

           #. a log is included at the current level in which case we index
              both the log and the parent transaction. Since the log could
              have been emitted from a transaction that was not originated
              from an indexed contract, we still index that transaction to
              be able to retrieve it later on.
           #. a transaction that was spooled is awaiting indexing to ensure
              it was included on-chain (contract origination,
              spooled contract call, etc).

        As such, no node requests should be made if there are no spooled
        transactions awaiting confirmation nor any events were emitted
        at the current level.
        """
        logs_at_level = list(filter(
            lambda log: log["blockNumber"] == level,
            self.logs,
        ))
        logs_tx_hash = [log["transactionHash"].hex() for log in logs_at_level]

        hashes_at_level = list(filter(
            lambda hash_tuple: hash_tuple[1] == level,
            self.hashes,
        ))
        hashes = [hash_tuple[0] for hash_tuple in hashes_at_level]

        if len(hashes):
            block = self.client.eth.get_block(level, True)
            for transaction in block.transactions:
                to = transaction.get('to', None)
                if to is None and transaction['hash'].hex() in hashes:
                    self.index_contract(level, transaction)
                elif (
                    to in self.addresses
                    or (transaction['hash'].hex() in logs_tx_hash and to)
                ):
                    self.index_call(level, transaction)

        for log in logs_at_level:
            if not log["removed"]:
                self.index_log(log)

        self.hashes = list(filter(
            lambda hash_tuple: hash_tuple[1] > level,
            self.hashes,
        ))

        self.logs = list(filter(
            lambda log: log['blockNumber'] != level,
            self.logs
        ))

    def get_contract_event_names(self, contract_abi, encoded_event_name):
        events = []
        for entry in contract_abi:
            if (
                "type" in entry
                and entry["type"] == "event"
                and "name" in entry
            ):
                candidate_encoded_event_name = "0x" + event_abi_to_log_topic(
                    entry,
                ).hex()
                if candidate_encoded_event_name == encoded_event_name:
                    events.append(entry["name"])

        return events

    def index_log(self, log):
        # In case of it's a log for the contract creation, right now it
        # will not be indexed by 'index_call' or 'index_contract' methods
        # so the transaction is not created
        # So right now I return if the tx not exists

        with db_transaction.atomic():
            transaction, created = self.transaction_class.objects.update_or_create(  # noqa: E501
                blockchain=self.blockchain,
                hash=log["transactionHash"].hex(),
                defaults=dict(
                    level=log["blockNumber"],
                )
            )
            transaction.state_set("done")

            if created:
                """
                If the transaction was created, then it's from the log of an
                indexed contract, we consider the original transaction to be
                a `function` kind which we don't index.

                This is crucial as it will otherwise by default be saved as
                an indexed contract. If so, then the indexer will try to
                index a contract with no address and raise!!

                Hence this must also be in the atomic block.
                """
                transaction.kind = "function"
                transaction.index = False
                transaction.save()

        contract = self.transaction_class.objects.filter(
            address=log["address"]
        ).first()
        contract_ci = self.client.eth.contract(
            address=contract.address,
            abi=contract.abi,
        )
        # first topic encodes event name
        encoded_log_name = log["topics"][0].hex()
        event_names = self.get_contract_event_names(
            contract_ci.abi,
            encoded_log_name,
        )
        for event_name in event_names:
            event = getattr(contract_ci.events, event_name)
            event_data = event().process_receipt({"logs": [log]})
            if event_data:
                self.event_class.objects.update_or_create(
                    name=event_name,
                    contract=contract,
                    transaction=transaction,
                    args=event_data[0]["args"],
                    event_index=event_data[0]["logIndex"],
                )

    def get_transactions_to_normalize(self):
        return self.transaction_class.objects.filter(
            transactionevent_set__normalized=False,
            caller=None,
            state='done',
        ).order_by(
            'created_at',
        ).distinct()



from decimal import Decimal
import logging
import requests
import urllib

from django.db.models import Q, signals
from django.core.exceptions import ValidationError
from pytezos.operation.result import OperationResult
from pytezos.rpc import RpcError

from djwebdapp.exceptions import PermanentError
from djwebdapp.models import Account, account_setup
from djwebdapp.provider import Provider

from djwebdapp_tezos.models import TezosTransaction

from pytezos import pytezos, Key


class TezosProvider(Provider):
    logger = logging.getLogger('djwebdapp_tezos')
    transaction_class = TezosTransaction

    def generate_secret_key(self):
        key = Key.generate(export=False)
        return key.public_key_hash(), key.secret_exponent

    def get_client(self, **kwargs):
        if self.wallet:
            kwargs['key'] = Key.from_secret_exponent(
                self.wallet.get_secret_key()
            )
        return pytezos.using(
            shell=self.blockchain.node_set.first().endpoint,
            **kwargs,
        )

    @property
    def head(self):
        return self.client.shell.head.metadata()['level_info']['level']

    def get_address(self):
        return self.client.key.public_key_hash()

    def get_balance(self, address=None):
        return Decimal(
            self.client.account(address or self.get_address())['balance']
        )

    def index_level(self, level):
        block = self.client.shell.blocks[level]
        for ops in block.operations():
            for number, op in enumerate(ops):
                for content in op.get('contents', []):
                    self.index_content(level, number, op, content)

    def index_content(self, level, number, op, content):
        self.logger.info(f'Indexing content {number}@{level} {op["hash"]}')
        # index content normally
        hash = op['hash']
        if (
            content['kind'] == 'origination'
            and 'metadata' in content
        ):
            result = content['metadata']['operation_result']
            address = result['originated_contracts'][0]
            if (
                address not in self.addresses
                and not self.check_hash(hash)
            ):
                # skip unknown contract originations
                return
            self.index_contract(level, op, content, number=number)
        elif (
            content['kind'] == 'transaction'
            and (
                content.get('destination', None) in self.addresses
                or self.check_hash(hash)
            )
        ):
            self.index_call(level, op, content, number=number)
        else:
            # index internal transactions if necessary
            for internal_op in OperationResult.iter_contents(content):
                if (
                    internal_op.get('destination', '') in self.addresses
                    or internal_op.get('source', '') in self.addresses
                ):
                    self.index_call(
                        level,
                        op,
                        content,
                        number=number,
                    )
                    break

    def index_contract(self, level, op, content, number):
        self.logger.info(f'Syncing origination {op["hash"]}')
        result = content['metadata']['operation_result']
        address = result['originated_contracts'][0]
        contract = self.transaction_class.objects.get(
            Q(address=address) | Q(hash=op['hash'])
        )
        contract.level = level
        contract.address = address
        contract.hash = op['hash']
        contract.gas = content['fee']
        contract.metadata = content
        contract.number = number
        contract.sender, _ = self.blockchain.account_set.get_or_create(
            address=op['contents'][0]['source'],
            blockchain=self.blockchain,
        )
        contract.state_set('done')

    def is_implicit_contract(self, address):
        return len(address) == 36 and address[:2] == 'tz'

    def index_origination(self, level, hash, content, caller=None,
                          number=None):
        self.logger.info(f'Syncing origination {hash}')

        originated_contracts = []
        for originated_address in content['result']['originated_contracts']:
            contract, created = self.transaction_class.objects.get_or_create(
                address=originated_address,
                blockchain=self.blockchain,
                caller=caller,
            )
            contract.level = level
            contract.hash = hash
            contract.gas = content.get('fee', 0)
            contract.metadata = content
            contract.nonce = content.get('nonce', -1)
            contract.sender = self.get_account(content['source'])
            contract.number = number
            contract.state_set('done')
            originated_contracts.append(contract)

        return originated_contracts

    def get_account(self, address):
        sender = Account.objects.filter(
            address=address,
            blockchain=self.blockchain,
        ).first()

        if not sender:
            sender = Account.objects.create(
                address=address,
                blockchain=self.blockchain,
                index=False,
            )

        return sender

    def index_transaction(self, level, hash, content, caller=None,
                          number=None):
        self.logger.info(f'Syncing transaction {hash}')

        counter = caller.counter if caller else content.get('counter', None)

        # figure destination contract
        destination_address = content['destination']

        if self.is_implicit_contract(destination_address):
            # this transaction targets an account
            contract = None
            receiver, _ = Account.objects.get_or_create(
                blockchain=self.blockchain,
                address=destination_address,
            )
            qs = receiver.transaction_received
        else:
            # this transaction targets a contract
            receiver = None
            contract = self.transaction_class.objects.filter(
                blockchain=self.blockchain,
                address=destination_address,
            ).first()
            if not contract:
                contract = self.transaction_class.objects.create(
                    blockchain=self.blockchain,
                    address=destination_address,
                    index=False,
                    number=number,
                )
            qs = contract.call_set.select_subclasses()

        call = qs.filter(
            hash=hash,
            nonce=content.get('nonce', -1),
            counter=counter,
            # we shouldn't take the level into account
            # when filtering due to confirm transactions
            # that "overide" the level in djwebdapp.provider.index
            # level=level,  <-- don't uncomment this
        ).first()

        if not call:
            call = self.transaction_class(
                hash=hash,
                counter=counter,
                contract=contract,
                receiver=receiver,
                blockchain=self.blockchain,
                nonce=content.get('nonce', -1),
                amount=int(content.get('amount', 0)),
                state='held',
                caller=caller,
                number=number,
                level=level,
            )

        # update call
        call.metadata = content
        call.sender = self.get_account(content['source'])

        # patch against empty args in pytezes
        if 'parameters' in content:
            call.function = content['parameters']['entrypoint']
            method = getattr(contract.interface, call.function)
            args = method.decode(call.metadata['parameters']['value'])
            if args == call.function:
                call.args = []
            else:
                if call.function not in args:
                    call.args = args
                else:
                    call.args = args[call.function]

        # save and return call
        call.state_set('done')

        return call

    def index_call(self, level, op, content, number=None):
        self.logger.info(f'Syncing transaction {op["hash"]}')

        operations = [op for op in OperationResult.iter_contents(content)]
        internal_operations = operations[1:]

        source = self.index_transaction(
            level,
            op['hash'],
            content,
            number=number,
        )

        internal_transactions = []
        for operation_content in internal_operations:
            if operation_content['kind'] == 'transaction':
                if 'destination' in operation_content:
                    internal_transaction = self.index_transaction(
                        level,
                        op['hash'],
                        operation_content,
                        source,
                        number=number,
                    )
                    internal_transactions.append(internal_transaction)

            if operation_content['kind'] == 'origination':
                if 'originated_contracts' in operation_content['result']:
                    internal_originations = self.index_origination(
                        level,
                        op['hash'],
                        operation_content,
                        source,
                        number=number,
                    )
                    internal_transactions += internal_originations

        return source

    def transfer(self, transaction):
        """ Execute a transfer transaction. """

        """
        rpc error if balance too low :

        RpcError ({'amount': '120000000000000000',
              'balance': '3998464237867',
              'contract': 'tz1gjaF81ZRRvdzjobyfVNsAeSC6PScjfQwN',
              'id': 'proto.006-PsCARTHA.contract.balance_too_low',
              'kind': 'temporary'},)
        """
        tx = self.client.transaction(
            destination=transaction.receiver.address,
            amount=transaction.amount,
        ).autofill().sign()
        result = self.write_transaction(tx, transaction)
        return result

    def deploy(self, transaction):
        if not self.client.balance():
            raise ValidationError(
                f'{transaction.sender.address} needs more than 0 tezies')
        elif not self.wallet.revealed:
            try:
                self.client.reveal().send(min_confirmations=1)
            except RpcError as exc:
                if len(exc.args) > 1:
                    raise

                if not exc.args[0]['id'].endswith('previously_revealed_key'):
                    raise
            self.wallet.revealed = True
            self.wallet.save()

        self.logger.debug(f'{transaction}.deploy(): start')
        if transaction.kind == 'contract':
            self.originate(transaction)
        elif transaction.kind == 'function':
            self.send(transaction)
        elif transaction.kind == 'transfer':
            self.transfer(transaction)
        else:
            transaction.error = f'Unknown transaction kind {transaction.kind}'
            transaction.state_set('failed')
            return

        transaction.sender.last_level = self.head
        transaction.sender.save()
        self.logger.info(f'{transaction}.deploy(): success')

    def originate(self, transaction):
        tx = self.client.origination(dict(
            code=transaction.micheline,
            storage=transaction.get_args(),
        )).autofill().sign()

        self.write_transaction(tx, transaction)

    def write_transaction(self, tx, transaction):
        origination = tx.inject(
            _async=False,
        )
        transaction.level = self.head + 1  # it'll be in the next block
        transaction.gas = origination['contents'][0]['fee']
        transaction.hash = origination['hash']
        transaction.counter = origination['contents'][0]['counter']
        transaction.save()

    def send(self, transaction):
        self.logger.debug(
            f'{transaction}: counter = {self.client.account()["counter"]}'
        )
        ci = self.client.contract(transaction.contract.address)
        method = getattr(ci, transaction.function)
        try:
            args = transaction.get_args()
            if isinstance(args, dict):
                tx = method(**args)
            else:
                tx = method(*args)
            if transaction.amount:
                tx = tx.with_amount(transaction.amount)
        except ValueError as e:
            raise PermanentError(*e.args)
        self.write_transaction(tx, transaction)

    def download(self, target):
        """
        Import transactions from tzkt for a contract
        """
        api = self.blockchain.configuration.get(
            'tzkt_url',
            'https://api.tzkt.io',  # default indexer?
        )
        url = f'{api}/v1/operations/transactions?'
        url += f'&target={urllib.parse.quote(target)}'
        url += f'&level.le={self.head - 1}'

        contract, _ = self.transaction_class.objects.get_or_create(
            blockchain=self.blockchain,
            address=target,
        )

        def get(offset, limit):
            return requests.get(url + f'&offset={offset}&limit={limit}').json()

        def yield_operations():
            offset = 0
            limit = 10_000
            while data := get(offset, limit):
                for operation in data:
                    if operation['type'] != 'transaction':
                        continue

                    if operation['status'] != 'applied':
                        continue

                    yield operation
                offset += limit

        total = 0
        operations = dict()
        for operation in yield_operations():
            total += 1

            # generate a key for the operation group
            key = (
                operation['level'],
                operation['hash'],
                operation['counter'],
                operation.get('nonce', 0),
            )

            operations[key] = operation

        # disable automatic refresh of balances to speed up the whole process
        signals.pre_save.disconnect(account_setup, sender=Account)
        # we're also going to cache Account objects
        accounts = dict()

        # fetch all calls we already have in DB
        calls = {
            (call.level, call.hash, call.counter, call.nonce): call
            for call in self.transaction_class.objects.filter(
                contract=contract,
                state__in=('confirm', 'done'),
            )
        }

        number = 0
        for key in sorted(operations.keys()):
            number += 1
            level, hash, counter, nonce = key
            self.logger.debug(' '.join((
                f'[{number}/{total}]',
                hash,
                f'level={level}',
                f'counter={counter}',
                f'nonce={nonce if isinstance(nonce, int) else -1}',
            )))

            if key in calls:
                continue  # let's not update for now

            operation = operations[key]

            if operation['sender']['address'] in accounts:
                sender = accounts[operation['sender']['address']]
            else:
                sender, _ = Account.objects.get_or_create(
                    address=operation['sender']['address'],
                    blockchain=self.blockchain,
                )
                accounts[sender.address] = sender

            function = operation['parameter']['entrypoint']
            args = operation['parameter']['value']

            for key, value in args.items():
                try:
                    args[key] = int(value)
                except ValueError:
                    continue

            call = self.transaction_class(
                blockchain=self.blockchain,
                contract=contract,
                sender=sender,
                level=level,
                hash=hash,
                counter=counter,
                nonce=nonce if isinstance(nonce, int) else -1,
                kind='call',
                function=function,
                args=args,
                metadata=operation,
                gas=operation['gasUsed'],
            )
            call.state_set('done')

        # reconnect Account signal
        signals.pre_save.connect(account_setup, sender=Account)



# Generated by Django 4.2.7 on 2023-11-28 17:45

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):
    dependencies = [
        ("djwebdapp", "0017_event"),
        ("djwebdapp_ethereum", "0004_ethereumevent"),
        ("djwebdapp_example_ethereum", "0001_initial"),
    ]

    operations = [
        migrations.CreateModel(
            name="FA12EthereumBalanceMovement",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("amount", models.PositiveIntegerField(default=0)),
                (
                    "account",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="djwebdapp.account",
                    ),
                ),
                (
                    "event",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="fa12_balance_movements",
                        to="djwebdapp_ethereum.ethereumevent",
                    ),
                ),
                (
                    "fa12",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="djwebdapp_example_ethereum.fa12ethereum",
                    ),
                ),
            ],
        ),
    ]



from djwebdapp.models import Account

# Given a blockchain and no address, djwebdapp will create a keypair
admin = Account.objects.create(blockchain=blockchain, name='Admin')

# Now, provision this wallet with some eths from the default client wallet
client.eth.send_transaction(dict(
    to=admin.address,
    value=client.to_wei(4_000_000, 'ether'),
))

# Wait until the account receives the balance
while not admin.balance:
    admin.refresh_balance()



# Generated by Django 4.1.1 on 2023-02-03 14:36

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("djwebdapp_tezos", "0003_tezoscall_tezoscontract"),
    ]

    operations = [
        migrations.DeleteModel(
            name="TezosCall",
        ),
        migrations.DeleteModel(
            name="TezosContract",
        ),
    ]



# Generated by Django 4.0.5 on 2022-06-27 13:20

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('djwebdapp', '0006_blockchain_min_confirmations_alter_transaction_state'),
    ]

    operations = [
        migrations.AlterField(
            model_name='transaction',
            name='amount',
            field=models.DecimalField(blank=True, db_index=True, decimal_places=18, default=0, max_digits=80),
        ),
    ]



from djwebdapp.models import Account
from djwebdapp.normalizers import Normalizer

from djwebdapp_example_tezos.models import FA12Tezos, FA12TezosMint, FA12TezosBalance


class FA12TezosNormalizer(Normalizer):
    def mint(self, call, contract):
        account, _ = Account.objects.get_or_create(
            address=call.args['_to'],
        )
        call, _ = FA12TezosMint.objects.update_or_create(
            tezostransaction_ptr_id=call.id,
            defaults=dict(
                target_contract=contract,
                mint_account=account,
                mint_amount=call.args['value'],
            )
        )
        balance, _ = FA12TezosBalance.objects.get_or_create(
            account=account,
            fa12=contract,
        )
        balance.balance += call.mint_amount
        balance.save()



# Generated by Django 4.1a1 on 2022-09-08 17:50

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ("djwebdapp_tezos", "0001_initial"),
    ]

    operations = [
        migrations.AddField(
            model_name="tezostransaction",
            name="caller",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="_internal_calls",
                to="djwebdapp_tezos.tezostransaction",
            ),
        ),
    ]



# Generated by Django 4.1.1 on 2022-09-23 10:11

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("djwebdapp", "0011_rename_max_level_blockchain_index_level"),
    ]

    operations = [
        migrations.AddField(
            model_name="transaction",
            name="number",
            field=models.IntegerField(blank=True, default=None, null=True),
        ),
    ]



# Use the transaction model with an amount argument to transfer coins
from djwebdapp.models import Transaction
transaction = Transaction.objects.create(
    name='Provision 1.000.000 coins',
    amount=1_000_000,
    sender=bootstrap,
    receiver=new_wallet,
    blockchain=blockchain,
)

# Deploy the transaction now
transaction.deploy()

# Wait for confirmation blocks
transaction.blockchain.wait()

# Index the transaction
blockchain.provider.index()

# And refresh balance
new_wallet.refresh_balance()



#!/usr/bin/env python3

from django.urls import path

from fa2.views import Fa2TokenRetrieveAPIView


urlpatterns = [
    path(
        "fa2/<str:token_address>/<int:token_id>/",
        Fa2TokenRetrieveAPIView.as_view(),
        name="fa2-detail",
    ),
]



from dataclasses import dataclass
from pymich.michelson_types import Address, BigMap, Contract, Nat
from pymich.stdlib import SENDER


@dataclass
class TransferParam:
    _from: Address
    _to: Address
    value: Nat


@dataclass
class Proxy(Contract):
    fa12: Address
    admin: Address

    def transfer(self, _from: Address, _to: Address, value: Nat):
        param = TransferParam(
            _from=_from,
            _to=_to,
            value=value,
        )
        transaction(Contract(self.fa12), param, TransferParam)
        # ideally, we'd have:
        # self.fa12 typed as FA12 and call:
        # self.fa12.transfer(_from, _to, value)

    def getAdmin(self) -> Address:  # noqa
        return self.admin



from django.db import models

from djwebdapp_tezos.models import TezosCall, TezosContract


class FA12Tezos(TezosContract):
    contract_name = 'FA12'
    normalizer_class = 'FA12TezosNormalizer'
    token_name = models.CharField(max_length=200)
    token_symbol = models.CharField(max_length=10)

    def get_init_storage(self):
        return dict(
            tokens={},
            total_supply=0,
            owner=self.sender.address,
        )


class FA12TezosMint(TezosCall):
    entrypoint = 'mint'
    target_contract = models.ForeignKey(
        'FA12Tezos',
        on_delete=models.CASCADE,
    )
    mint_account = models.ForeignKey(
        'djwebdapp.Account',
        on_delete=models.CASCADE,
    )
    mint_amount = models.PositiveIntegerField()

    def __str__(self):
        return f'mint({self.mint_account.address}, {self.mint_amount})'

    def get_args(self):
        return dict(
            _to=self.mint_account.address,
            value=self.mint_amount,
        )


class FA12TezosBalance(models.Model):
    fa12 = models.ForeignKey(
        'FA12Tezos',
        on_delete=models.CASCADE,
    )
    account = models.ForeignKey(
        'djwebdapp.Account',
        on_delete=models.CASCADE,
    )
    balance = models.PositiveIntegerField(default=0)

    class Meta:
        unique_together = (
            ('fa12', 'account'),
        )

    def __str__(self):
        return f'{self.account} balance: {self.balance}'





class BaseException(Exception):
    """ Base exception class for all exceptions of this package. """
    pass


class PermanentError(BaseException):
    """ Raised when a transaction can not be deployed. """


class TemporaryError(BaseException):
    """ Raised when a transaction has not been deployed. """



# Generated by Django 4.0.5 on 2022-06-17 13:27

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('djwebdapp', '0005_alter_transaction_unique_together_and_more'),
    ]

    operations = [
        migrations.AddField(
            model_name='blockchain',
            name='min_confirmations',
            field=models.PositiveIntegerField(default=2, help_text='Number of confirmation blocks from confirm to done states'),
        ),
        migrations.AlterField(
            model_name='transaction',
            name='state',
            field=models.CharField(choices=[('held', 'Held'), ('deleted', 'Deleted during reorg'), ('aborted', 'Aborted'), ('deploy', 'To deploy'), ('deploying', 'Deploying'), ('retry', 'To retry'), ('retrying', 'Retrying'), ('confirm', 'Deployed to confirm'), ('done', 'Confirmed finished')], db_index=True, default='held', max_length=200),
        ),
    ]



# Generated by Django 4.2.1 on 2023-05-11 16:23

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("djwebdapp", "0015_alter_transaction_state"),
    ]

    operations = [
        migrations.AlterField(
            model_name="transaction",
            name="has_code",
            field=models.BooleanField(
                default=False,
                help_text="Checked if this transaction has smart contract code.",
            ),
        ),
    ]



# Generated by Django 4.1.1 on 2023-01-20 16:54

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("djwebdapp_tezos", "0002_tezostransaction_caller"),
    ]

    operations = [
        migrations.CreateModel(
            name="TezosCall",
            fields=[],
            options={
                "proxy": True,
                "indexes": [],
                "constraints": [],
            },
            bases=("djwebdapp_tezos.tezostransaction",),
        ),
        migrations.CreateModel(
            name="TezosContract",
            fields=[],
            options={
                "proxy": True,
                "indexes": [],
                "constraints": [],
            },
            bases=("djwebdapp_tezos.tezostransaction",),
        ),
    ]



# Generated by Django 5.2 on 2025-04-15 17:04

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('djwebdapp', '0017_event'),
    ]

    operations = [
        migrations.AddField(
            model_name='blockchain',
            name='chain_id',
            field=models.IntegerField(blank=True, null=True),
        ),
    ]



# Generated by Django 4.1.1 on 2022-09-11 17:24

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("djwebdapp", "0009_account_index"),
    ]

    operations = [
        migrations.AlterField(
            model_name="transaction",
            name="nonce",
            field=models.IntegerField(default=-1),
        ),
    ]



from django.db import models

from djwebdapp_ethereum.models import (
    EthereumCall,
    EthereumContract,
    EthereumEvent,
)


class FA12Ethereum(EthereumContract):
    contract_name = 'FA12'
    normalizer_class = 'FA12EthereumNormalizer'

    token_name = models.CharField(max_length=200)
    token_symbol = models.CharField(max_length=10)

    def get_args(self):
        return (self.token_name, self.token_symbol)


class FA12EthereumMint(EthereumCall):
    entrypoint = 'mint'
    target_contract = models.ForeignKey(
        'FA12Ethereum',
        on_delete=models.CASCADE,
    )
    mint_account = models.ForeignKey(
        'djwebdapp.Account',
        on_delete=models.CASCADE,
    )
    mint_amount = models.PositiveIntegerField()

    def __str__(self):
        return f'mint({self.mint_account.address}, {self.mint_amount})'

    def get_args(self):
        return (self.mint_account.address, self.mint_amount)


class FA12EthereumBalanceMovement(models.Model):
    event = models.ForeignKey(
        EthereumEvent,
        on_delete=models.CASCADE,
        related_name='fa12_balance_movements',
    )
    fa12 = models.ForeignKey(
        'FA12Ethereum',
        on_delete=models.CASCADE,
    )
    account = models.ForeignKey(
        'djwebdapp.Account',
        on_delete=models.CASCADE,
    )
    amount = models.PositiveIntegerField(default=0)

    def save(self, *args, **kwargs):
        super().save(*args, **kwargs)
        total_balance = FA12EthereumBalanceMovement.objects.filter(
            account=self.account,
            fa12=self.fa12,
        ).aggregate(
            balance=models.Sum('amount'),
        )['balance']

        FA12EthereumBalance.objects.update_or_create(
            account=self.account,
            fa12=self.fa12,
            defaults=dict(
                balance=total_balance,
            )
        )

    def delete(self, *args, **kwargs):
        super().delete(*args, **kwargs)
        total_balance = FA12EthereumBalanceMovement.objects.filter(
            account=self.account,
            fa12=self.fa12,
        ).aggregate(
            balance=models.Sum('amount'),
        )['balance']

        FA12EthereumBalance.objects.update_or_create(
            account=self.account,
            fa12=self.fa12,
            defaults=dict(
                balance=total_balance,
            )
        )


class FA12EthereumBalance(models.Model):
    fa12 = models.ForeignKey(
        'FA12Ethereum',
        on_delete=models.CASCADE,
    )
    account = models.ForeignKey(
        'djwebdapp.Account',
        on_delete=models.CASCADE,
    )
    balance = models.PositiveIntegerField(default=0)

    class Meta:
        unique_together = (
            ('fa12', 'account'),
        )

    def __str__(self):
        return f'{self.account} balance: {self.balance}'



from multiprocessing import get_context
import logging
import random

from django import db
from django.db.models import Q

from djwebdapp.models import Event, Transaction


def call_deploy(arg):
    """
    Wrap a call in a try/except with logging.

    :param arg: Tuple of (logger, call)
    """
    logger, call = arg

    logger.debug(f'starting {call} ...')
    try:
        call.deploy()
    except Exception:
        logger.exception(f'failed {call}')
    else:
        logger.info(f'success {call}')

    return call


def get_calls_distinct_sender(calls_query_set, n_calls):
    """
    Given a QS of calls, return a list of calls with distinct senders.

    :param calls_query_set: Queryset of Calls
    :param n_calls: Number of calls to return
    """
    senders = set()
    distinct_calls = []
    for call in calls_query_set:
        if call.sender not in senders and len(distinct_calls) < n_calls:
            distinct_calls.append(call)
            senders.add(call.sender)

    return distinct_calls


class Provider:
    """
    Base Provider class, encapsulates business logic with blockchains.

    Instanciating a Provider requires either a `blockchain` or a `wallet`
    argument.

    .. warning:: Do not use this class directly, it is meant to be sub-classed
                 for each blockchain type.

    :param blockchain: :py:class:`~djwebdapp.models.Blockchain` object to
                       instanciate the Provider with, in which case the Python
                       client for the blockchain will use its default account.

    :param wallet: :py:class:`~djwebdapp.models.Account` object to bind the
                    provider with, in which case the
                    :py:attr:`~djwebdapp.models.Account.secret_key` must be set
                    in the Account.

    .. note:: You don't need to pass both `wallet` and `blockchain` arguments,
              the provider will instanciate with the `blockchain` attribute of
              the `wallet` argument if any.

    .. py:attribute:: blockchain

        :py:class:`~djwebdapp.models.Blockchain` this provider was instanciated
        for.

    .. py:attribute:: wallet

        :py:class:`~djwebdapp.models.Account` this provider was instanciated
        for, if any.

        .. danger:: When instanciating a provider with a :py:attr:`wallet`,
                    make sure the :py:class:`~djwebdapp.models.Account` has a
                    secret key! Otherwise, the blockchain client will not be
                    able to send any transaction!

    .. py:attribute:: transaction_class

        :py:class:`~djwebdapp.models.Transaction` subclass this provider works
        with. For example, :py:class:`~djwebdapp_tezos.models.TezosTransaction`
        if it's a tezos provider, or a
        :py:class:`~djwebdapp_ethereum.models.EthereumTransaction` if it's an
        ethereum provider. This is supposed to be set as a class attribute by
        the provider subclass.

    .. py:attribute:: exclude_states

        States to exclude when searching for transactions to deploy.
    """
    exclude_states = (
        'held', 'aborted', 'import', 'importing', 'confirm', 'done'
    )

    def __init__(self, blockchain=None, wallet=None):
        self.wallet = wallet
        self.blockchain = wallet.blockchain if wallet else blockchain

    def generate_secret_key(self):
        """
        Generate a secret key.

        Raises NotImplemented in base Provider class.
        """
        raise NotImplementedError()

    @property
    def head(self):
        """
        Return the current block number.

        Raises NotImplemented in base Provider class.
        """
        raise NotImplementedError()

    def download(self, target: str):
        """
        Download a contract history from the configured indexer.

        This will use an indexer such as etherscan or tzkt to download the
        history from a web-API rather than by indexing the blockchain which
        could take a really long while for large contracts.

        :param target: String address of the contract to download history for.
        """
        raise NotImplementedError()

    def index_level(self, level: int):
        """
        Index a given block level.

        Left to implement in new provider subclasses.
        """
        raise NotImplementedError()

    def get_client(self, wallet=None):
        """
        Return the Python client that provider encapsulates.

        :param wallet: :py:class:`~djwebdapp.models.Account` object to use,
                        note that it *must* have a
                        :py:attr:`~djwebdapp.models.Account.secret_key`.
        """
        raise NotImplementedError()

    @property
    def client(self):
        """
        Cached result of :py:meth:`get_client()`
        """
        cached = getattr(self, '_client', None)
        if cached:
            return cached
        self._client = self.get_client()
        return self._client

    @client.setter
    def client(self, value):
        self._client = value

    def index_init(self):
        """
        Query the database for transactions hashes and contract addresses to
        index in :py:meth:`index()`

        Provisions
        """
        self.hashes = list(self.transaction_class.objects.filter(
            blockchain=self.blockchain
        ).filter(
            Q(state='confirm') | ~Q(hash=None)
        ).exclude(
            state='done'
        ).values_list('hash', 'level'))

        self.logger.info(f'Found {len(self.hashes)} transactions to index')

        self.contracts = self.transaction_class.objects.filter(
            blockchain=self.blockchain,
            index=True,
            kind='contract',
        ).filter(
            Q(state='confirm') | ~Q(address=None),
        )
        print(f'Found {len(self.contracts)} contracts to index')

        self.addresses = self.contracts.values_list(
            'address',
            flat=True,
        )
        self.logger.info(f'Found {len(self.addresses)} addresses to index')

    def check_hash(self, hash_to_check):
        """
        Returns True if hash is in self.hashes
        """
        return any(hash_to_check == hash for hash, _ in self.hashes)

    def deploy(self, transaction):
        """
        Deploy a given :py:class:`~djwebdapp.models.Transaction` object.
        """
        raise NotImplementedError()

    def reorg(self):
        """
        Handle reorg if necessary.

        Compare the head level with the last indexed level, if it's superior
        then consider a reorg happened on the blockchain.

        In this case, empty the level, hash, and address of every
        :py:class:`~djwebdapp.models.Transaction` in DB which has a level
        greater than or equal to the current head level, and set their state to
        `deleted`.
        """
        current_level = self.head
        reorg = (
            self.blockchain.index_level
            and current_level < self.blockchain.index_level
        )
        if reorg:
            self.logger.warning(
                f'Detected reorg in {self.blockchain} '
                f'from {self.blockchain.index_level} to {current_level}'
            )
            # reorg
            Transaction.objects.filter(
                sender__blockchain=self.blockchain,
                level__gte=current_level,
            ).exclude(level=None).update(
                level=None,
                hash=None,
                address=None,
                state='deleted',
            )
            self.blockchain.index_level = current_level
            self.blockchain.save()
            return True  # commit to reorg in a transaction

    def index(self):
        """
        Index the blockchain.

        Return if :py:meth:`reorg()` returns True.

        Iterate over each level from the last indexed level in the
        :py:attr:`djwebdapp.models.Blockchain.index_level` column (or 0) up to
        the current head level. Call :py:meth:`index_level()` for each level.
        """
        if self.reorg():
            return  # commit to reorg in a transaction

        self.index_init()

        from django.db.models import Min
        level = self.blockchain.transaction_set.filter(
            state='confirm'
        ).aggregate(
            Min('level')
        )['level__min']
        if not level:
            if self.blockchain.index_level:
                level = self.blockchain.index_level
            else:
                level = self.blockchain.index_level = 0

        while level <= self.head:
            self.logger.info(f'Indexing level {level}')
            self.index_level(level)
            self.blockchain.index_level = level
            level += 1
        self.blockchain.save()

    def spool_contracts(self):
        """
        Return the contracts to deploy, used by :py:meth:`~spool()`.

        Return a QuerySet of :py:attr:`transaction_class` objects of this
        :py:attr:`blockchain` which:

        - are of :py:attr:`~djwebdapp.models.Transaction.kind` ``contract``
        - have no :py:attr:`~djwebdapp.models.Transaction.hash`
        - have code according to
          :py:attr:`~djwebdapp.models.Transaction.has_code`, so that we can
          actually deploy it
        - have no :py:attr:`~djwebdapp.models.Transaction.address`
        - which :py:attr:`~djwebdapp.models.Transaction.state` is not in
          :py:attr:`exclude_states`,
        - which sender :py:class:`~djwebdapp.models.Account` have not deployed
          to the blockchain during this level according to :py:attr:`head` and
          :py:attr:`djwebdapp.models.Account.last_level`
        - which sender has balance above 0
        - ordered by :py:attr:`~djwebdapp.models.Transaction.created_at`
          ascending order, so that it gets the oldest first.
        """
        return self.transaction_class.objects.filter(
            blockchain=self.blockchain,
            kind='contract',
            hash=None,
            sender__blockchain__is_active=True,
            has_code=True,
        ).filter(
            Q(address='')
            | Q(address=None)
        ).filter(
            Q(sender__last_level__lt=self.head)
            | Q(sender__last_level=None)
        ).exclude(
            Q(sender__balance=None)
            | Q(sender__balance=0)
            | Q(state__in=self.exclude_states)
        ).select_related(
            'blockchain'
        ).order_by(
            'created_at'
        ).select_subclasses()

    def spool_calls(self):
        """
        Return the calls to deploy in :py:meth:`~spool()`.

        Return a QuerySet of :py:attr:`transaction_class` objects of this
        :py:attr:`blockchain` which:

        - are of :py:attr:`~djwebdapp.models.Transaction.kind` ``function``
        - have no :py:attr:`~djwebdapp.models.Transaction.hash`
        - are related to a contract which does have an
          :py:attr:`~djwebdapp.models.Transaction.address`
        - which :py:attr:`~djwebdapp.models.Transaction.state` is not in
          :py:attr:`exclude_states`,
        - which sender :py:class:`~djwebdapp.models.Account` have not deployed
          to the blockchain during this level according to :py:attr:`~head` and
          :py:attr:`djwebdapp.models.Account.last_level`
        - which sender has balance above 0
        - ordered by :py:attr:`~djwebdapp.models.Transaction.created_at`
          ascending order, so that it gets the oldest first.
        """
        return self.transaction_class.objects.filter(
            blockchain=self.blockchain,
            kind='function',
            hash=None,
            sender__blockchain__is_active=True,
        ).filter(
            Q(sender__last_level__lt=self.head)
            | Q(sender__last_level__isnull=True)
        ).exclude(
            Q(sender__balance=None)
            | Q(sender__balance=0)
            | Q(state__in=self.exclude_states)
            | Q(contract__address='')
            | Q(contract__address__isnull=True)
        ).select_related(
            'blockchain'
        ).order_by(
            'created_at'
        ).select_subclasses()

    def spool_transfers(self):
        """
        Return the transfers to deploy, used by :py:meth:`~spool()`.

        Return a QuerySet of :py:attr:`transaction_class` objects of this
        :py:attr:`blockchain` which:

        - are of :py:attr:`~djwebdapp.models.Transaction.kind` ``transfer``
        - have no :py:attr:`~djwebdapp.models.Transaction.hash`
        - which :py:attr:`~djwebdapp.models.Transaction.state` is not in
          :py:attr:`exclude_states`,
        - which sender :py:class:`~djwebdapp.models.Account` have not deployed
          to the blockchain during this level according to :py:attr:`head` and
          :py:attr:`djwebdapp.models.Account.last_level`
        - which sender has balance above 0
        - ordered by :py:attr:`~djwebdapp.models.Transaction.created_at`
          ascending order, so that it gets the oldest first.
        """
        return self.transaction_class.objects.filter(
            blockchain=self.blockchain,
            kind='transfer',
            hash=None,
            sender__blockchain__is_active=True,
            sender__last_level__lt=self.head,
        ).exclude(
            Q(sender__balance=None)
            | Q(sender__balance=0)
            | Q(state__in=self.exclude_states)
        ).select_related(
            'blockchain'
        ).order_by(
            'created_at'
        )

    def spool(self):
        """
        Deploy the next transaction of any kind and return it.

        It checks for the next transaction with the following logic:

        - :py:meth:`~spool_transfers()`: is there any new transfer to deploy?
        - :py:meth:`~spool_contracts()`: is there any *new* contract to deploy
          from an account with balance?
        """
        # senders which have already deployed during this block must be
        # excluded
        # is there any new transfer to deploy from an account with balance?
        transfer = self.spool_transfers().filter(last_fail=None).first()
        if transfer:
            self.logger.info(f'Deploying transfer {transfer}')
            return transfer.deploy()
        self.logger.info('Found 0 transfers to deploy')

        # is there any new contract to deploy from an account with balance?
        contract = self.spool_contracts().filter(last_fail=None).first()

        if contract:
            dependency = contract.dependency_get()
            if dependency:
                contract = dependency
            self.logger.info(f'Deploying contract {contract}')
            contract.deploy()
            return contract
        self.logger.info('Found 0 contracts to deploy')

        n_calls = 15
        calls = self.spool_calls().filter(last_fail=None)
        distinct_calls = get_calls_distinct_sender(calls, n_calls)

        if distinct_calls:
            db.connections.close_all()
            pool = get_context("fork").Pool(n_calls)
            results = pool.map(
                call_deploy,
                [(self.logger, call) for call in list(distinct_calls)]
            )
            for result in results:
                result.save()

            if len(distinct_calls) == 1:
                return distinct_calls[0]
            else:
                return distinct_calls
        self.logger.info('Found 0 calls to send')

        # is there any transfer to retry from an account with balance?
        transfer = self.spool_transfers().order_by('last_fail').first()
        if transfer:
            self.logger.info(f'Retrying transfer {transfer}')
            transfer.deploy()
            return transfer
        self.logger.info('Found 0 transfer to retry')

        # any contract to retry?
        contract = self.spool_contracts().order_by('last_fail').first()
        if contract:
            self.logger.info(f'Retrying contract {contract}')
            contract.deploy()
            return contract
        self.logger.info('Found 0 contract to retry')

        # any call to retry?
        call = self.spool_calls().order_by('last_fail').first()
        if call:
            self.logger.info(f'Retrying function {call}')
            call.deploy()
            return call
        self.logger.info('Found 0 call to retry')

    def get_transactions_to_normalize(self):
        return self.transaction_class.objects.filter(
            normalized=False,
            caller=None,
            state='done',
        ).order_by(
            'created_at',
        )

    def normalize(self):
        """
        Run `normalize()` on all un-normalized transactions.

        Call
        :py:meth:`~djwebdapp.models.Transaction.normalize()`
        on each transaction that has not been normalized.

        Internal transactions are normalized after their caller is normalized.
        """
        def normalize_internal(transaction):
            if not transaction.normalized:
                return

            internal_calls_qs = transaction._internal_calls.order_by("nonce")
            for internal in internal_calls_qs.all():
                internal.normalize()

        def normalize_event(event):
            if event.normalized:
                return

            event.normalize()

        transactions = self.get_transactions_to_normalize()

        for transaction in transactions:
            transaction.normalize()
            normalize_internal(transaction)
            events_qs = transaction.transactionevent_set.order_by(
                "event_index",
            ).all()
            for event in events_qs:
                event_subclass = Event.objects.get_subclass(pk=event.pk)
                normalize_event(event_subclass)

    def get_balance(self, address=None):
        """
        Query the blockchain and return the balance of an address.

        .. py:attribute:: address

            Address to get the balance of, use the current client address by
            default.
        """
        raise NotImplementedError()


def fakehash(leet):
    return f'0x{leet}5EF2D798D17e2ecB37' + str(random.randint(
        1000000000000000, 9999999999999999
    ))


class Success(Provider):
    logger = logging.getLogger('djwebdapp_test')

    def get_balance(self, address=None):
        return 1_000_000



# Generated by Django 5.2 on 2025-11-20 11:43

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("djwebdapp", "0019_alter_event_event_index_alter_event_normalized_and_more"),
    ]

    operations = [
        migrations.AlterField(
            model_name="account",
            name="address",
            field=models.CharField(
                blank=True, db_index=True, max_length=255, null=True
            ),
        ),
    ]



import json
import os

import dateutil.parser

from pytezos import ContractInterface

from django.db import models
from django.db.models import signals
from django.dispatch import receiver

from djwebdapp.models import Transaction
from djwebdapp.normalizers import Normalizer


class TezosTransaction(Transaction):
    """
    Base class for tezos transactions.

    .. py:attribute:: micheline

        Smart contract micheline JSON code.
    """
    unit_smallest = 'xTZ'
    contract = models.ForeignKey(
        'self',
        on_delete=models.CASCADE,
        related_name='call_set',
        null=True,
        blank=True,
        help_text='Smart contract, appliable to method call',
    )
    micheline = models.JSONField(
        default=dict,
        blank=True,
        null=True,
        help_text='Smart contract Micheline, if this is a smart contract',
    )
    caller = models.ForeignKey(
        "TezosTransaction",
        on_delete=models.CASCADE,
        null=True,
        related_name="_internal_calls",
    )

    def save(self, *args, **kwargs):
        """
        Set :py:attr:`~djwebdapp.models.Transaction.has_code` if
        :py:attr:`~micheline`.
        """
        if self.micheline:
            self.has_code = True
        return super().save(*args, **kwargs)

    @property
    def interface(self):
        return ContractInterface.from_micheline(self.micheline)

    @property
    def is_internal(self):
        return bool(self.caller_id)

    @property
    def internal_calls(self):
        if self.is_internal:
            txgroup_internal_calls_qs = self.caller._internal_calls
        else:
            txgroup_internal_calls_qs = self._internal_calls

        tx_internal_calls_qs = txgroup_internal_calls_qs.filter(
            nonce__gte=self.nonce if self.nonce else 0,
            sender__address=self.contract.address,
        )
        return tx_internal_calls_qs.order_by("nonce").all()

    @property
    def storage(self):
        from pytezos.operation.result import OperationResult
        if self.metadata:
            tx_op = OperationResult.from_transaction(self.metadata)
            contract = self.contract_subclass()
            if contract:
                return contract.interface.storage.decode(tx_op.storage)

    @property
    def timestamp(self):
        client = self.provider.client
        timestamp_str = client.shell.blocks[self.level].header.shell()
        return dateutil.parser.isoparse(timestamp_str["timestamp"])


@receiver(signals.pre_save, sender=TezosTransaction)
def contract_micheline(sender, instance, **kwargs):
    if not instance.address or instance.micheline:
        return

    if instance.kind != 'contract':
        return  # only fetch micheline for contracts

    interface = instance.blockchain.provider.client.contract(instance.address)
    instance.micheline = interface.to_micheline()


class TezosContract(TezosTransaction):
    contract_name = None
    normalizer_class = Normalizer

    class Meta:
        proxy = True

    def save(self, *args, **kwargs):
        """
        Set :py:attr:`~djwebdapp_tezos.models.TezosTransaction.micheline` if
        :py:attr:`~djwebdapp.models.Transaction.contract_name` is set.
        """
        if self.contract_name and not self.micheline:
            self.micheline = self.get_contract_interface().to_micheline()
        return super().save(*args, **kwargs)

    def get_contract_interface(self):
        if self.micheline:
            return ContractInterface.from_micheline(self.micheline)
        elif os.path.exists(f'{self.contract_path}.json'):
            with open(f'{self.contract_path}.json') as micheline:
                return ContractInterface.from_micheline(
                    json.loads(micheline.read())
                )
        elif os.path.exists(f'{self.contract_path}.tz'):
            with open(f'{self.contract_path}.tz') as michelson:
                return ContractInterface.from_michelson(michelson.read())
        raise Exception('Contract sources could not be found')

    def get_init_storage(self):
        raise NotImplementedError

    def get_michelson_storage(self):
        contract_interface = self.get_contract_interface()
        return contract_interface.storage.encode(self.get_init_storage())

    def get_args(self):
        return self.get_michelson_storage()


class TezosCall(TezosTransaction):
    entrypoint = None
    normalizer_class = None
    target_contract = None

    class Meta:
        proxy = True



from django.db import models
from djwebdapp.models import Account
from djwebdapp_multisig.models import MultisigContract

from djwebdapp_tezos.models import TezosCall, TezosContract


class MultisigedAbstractContract(TezosContract):
    multisig = models.ForeignKey(
        MultisigContract,
        on_delete=models.SET_NULL,
        null=True,
    )

    class Meta:
        abstract = True


class Fa2Contract(MultisigedAbstractContract):
    contract_name = "fa2"
    normalizer_class = 'Fa2Normalizer'
    manager = models.ForeignKey(
        Account,
        on_delete=models.CASCADE,
        related_name="fa2_administered",
        null=True,
    )
    metadata_uri = models.CharField(max_length=500)

    def get_init_storage(self):
        init_storage = self.get_contract_interface().storage.dummy()
        init_storage["manager"] = self.manager.address
        init_storage["multisig"] = self.multisig.address
        init_storage["metadata"] = {
            "": self.metadata_uri.encode(),
        }
        init_storage["ledger"] = {}

        return init_storage


class Fa2Token(models.Model):
    token_id = models.IntegerField()
    metadata_uri = models.CharField(max_length=500)
    contract = models.ForeignKey(
        Fa2Contract,
        on_delete=models.CASCADE,
    )

    class Meta:
        unique_together = ("token_id", "contract")


class MintCall(TezosCall):
    entrypoint = "mint"
    owner = models.ForeignKey(
        Account,
        on_delete=models.CASCADE,
        related_name="mintcall_owner_set",
    )
    token_id = models.PositiveBigIntegerField()
    token_amount = models.PositiveBigIntegerField()
    metadata_uri = models.CharField(max_length=500)

    def get_args(self):
        return {
            "token_id": self.token_id,
            "amount_": self.token_amount,
            "owner": self.owner.address,
            "token_metadata": {
                "": self.metadata_uri.encode(),
            },
        }


class UpdateProxyCall(TezosCall):
    entrypoint = "updateProxy"
    remove_proxy = models.BooleanField(default=False)
    proxy = models.ForeignKey(
        Account,
        on_delete=models.CASCADE,
        related_name="fa2contract_proxy_set",
    )

    def get_args(self):
        variant_type = "add_proxy"
        if self.remove_proxy:
            variant_type = "remove_proxy"

        return {variant_type: self.proxy.address}


class Balance(models.Model):
    token = models.ForeignKey(
        Fa2Token,
        on_delete=models.CASCADE,
    )
    account = models.ForeignKey(
        Account,
        related_name="fa2_balances",
        on_delete=models.CASCADE,
    )
    amount = models.PositiveBigIntegerField(default=0)

    class Meta:
        unique_together = ("token", "account")



import pytest

from djwebdapp.models import Transaction
from djwebdapp.normalizers import Normalizer


def test_str_name():
    assert str(Transaction(name='test')) == 'test'


def test_str_contract():
    assert str(Transaction(kind='contract', address='test')) == 'test'


def test_str_hash():
    assert str(Transaction(hash='test')) == 'test'


def test_str_function():
    assert str(Transaction(function='test')) == 'test()'


def test_str_amount():
    assert str(Transaction(amount=2)) == '2'
    Transaction.unit_smallest = 'XX'
    assert str(Transaction(amount=2)) == '2XX'
    Transaction.unit_smallest = None


def test_str_pk():
    assert str(Transaction(pk=3)) == '3'


@pytest.fixture
def transaction_normalizer():
    class FailNormalizer(Normalizer):
        fail = False
        @classmethod
        def deploy(cls, transaction, contract):
            if cls.fail:
                raise Exception('FailNormalizer')

    Transaction.normalizer_class = FailNormalizer
    yield Transaction
    Transaction.normalizer_class = None


@pytest.mark.django_db
def test_normalize_error(account, blockchain, transaction_normalizer):
    tx = transaction_normalizer(
        blockchain=blockchain,
        kind='contract',
        sender=account,
    )
    transaction_normalizer.normalizer_class.fail = True
    tx.save()
    tx.normalize()
    assert not tx.normalized
    assert tx.error.startswith('Traceback (most recent call')
    assert tx.error.endswith('Exception: FailNormalizer\n')
    assert tx.last_fail


@pytest.mark.django_db
def test_normalize_success(account, blockchain, transaction_normalizer):
    tx = transaction_normalizer(
        blockchain=blockchain,
        kind='contract',
        sender=account,
    )
    tx.save()
    tx.normalize()
    assert tx.normalized
    assert not tx.error
    assert not tx.last_fail



# Generated by Django 4.1.5 on 2023-02-15 08:40

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("djwebdapp", "0014_alter_account_unique_together"),
    ]

    operations = [
        migrations.AlterField(
            model_name="transaction",
            name="state",
            field=models.CharField(
                choices=[
                    ("held", "Held"),
                    ("deleted", "Deleted during reorg"),
                    ("aborted", "Aborted"),
                    ("deploy", "To deploy"),
                    ("deploying", "Deploying"),
                    ("retry", "To retry"),
                    ("retrying", "Retrying"),
                    ("confirm", "Deployed to confirm"),
                    ("done", "Confirmed finished"),
                ],
                db_index=True,
                default="deploy",
                max_length=200,
            ),
        ),
    ]



import binascii
import datetime
import importlib
import networkx
import os
import time
import traceback
import uuid

from django.conf import settings
from django.db import models
from django.db.models import Q, Max, signals
from django.dispatch import receiver
from django.utils import timezone
from django.utils.translation import gettext_lazy as _
from picklefield.fields import PickledObjectField

from fernet_fields import EncryptedTextField
from model_utils.managers import InheritanceManager


SETTINGS = dict(
    PROVIDERS=(
        ('djwebdapp_tezos.provider.TezosProvider', 'Tezos'),
        ('djwebdapp_ethereum.provider.EthereumProvider', 'Ethereum'),
        ('djwebdapp.provider.Success', 'Test that always succeeds'),
        ('djwebdapp.provider.FailDeploy', 'Test that fails deploy'),
        ('djwebdapp.provider.FailWatch', 'Test that fails watch'),
    )
)
SETTINGS.update(getattr(settings, 'DJBLOCKCHAIN', {}))


class Account(models.Model):
    """
    A blockchain account.

    .. py:attribute:: name

        Optionnal name for this account.

    .. py:attribute:: description

        Optionnal description for this account.

    .. py:attribute:: created_at

        Automatic datetime of the creation of this account in the database.

    .. py:attribute:: updated_at

        Automatic datetime of the last update of this account in the database.

    .. py:attribute:: address

        Account address on the blockchain.

    .. py:attribute:: blockchain

        Foreign key to the :py:class:`~Blockchain` model this account is
        related to.

    .. py:attribute:: balance

        Decimal balance of this account, updated by the
        :py:meth:`~refresh_balance()` method. with the `refresh_balances`
        management command. Uses

    .. py:attribute:: owner

        Foreign key to your User model which owns this account, if any.

    .. py:attribute:: secret_key

        djfernet encrypted private key of this account.

        This is used to sign transactions. The secret key **must** be set with
        the :py:meth:`~set_secret_key()` method.

        A secret key is generated automatically by the
        :py:func:`account_setup()` function *if* the account is not created
        with a :py:attr:`address`.

    .. py:attribute:: counter

        Counter of transactions sent from this account.

    .. py:attribute:: last_level

        Last block level when a transaction was sent from this account.

    .. py:attribute:: index

        Boolean to indicate wether the indexer should index
        all transactions or not.
    """

    name = models.CharField(
        max_length=100,
        null=True,
        blank=True,
    )
    description = models.TextField(
        null=True,
        blank=True,
    )
    created_at = models.DateTimeField(
        null=True,
        blank=True,
        auto_now_add=True,
    )
    updated_at = models.DateTimeField(
        null=True,
        blank=True,
        auto_now=True,
    )
    address = models.CharField(
        max_length=255,
        blank=True,
        null=True,
        db_index=True,
    )
    blockchain = models.ForeignKey(
        'Blockchain',
        on_delete=models.CASCADE,
    )
    balance = models.DecimalField(
        max_digits=80,
        decimal_places=18,
        blank=True,
        editable=False,
        default=0,
    )
    owner = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        null=True,
        blank=True,
        on_delete=models.SET_NULL,
    )
    secret_key = EncryptedTextField()
    revealed = models.BooleanField(default=False)
    counter = models.PositiveIntegerField(null=True)
    last_level = models.PositiveIntegerField(null=True)
    index = models.BooleanField(
        default=True,
        help_text='Wether the indexer should index all transactions or not',
    )

    objects = InheritanceManager()

    class Meta:
        unique_together = (
            ('blockchain', 'address'),
        )

    def __str__(self):
        """ Return given name or address or id. """
        return self.name or self.address or self.id

    @property
    def provider(self):
        """
        Return a Provider instance for this wallet.

        The :py:attr:`~Blockchain.provider_cls` returns the class that will be
        used.
        """
        return self.blockchain.provider_cls(wallet=self)

    def refresh_balance(self, commit=True):
        """
        Refresh the balance of this account.

        Using :py:meth:`~djwebdapp.provider.Provider.get_balance()`, it
        will update :py:attr:`~balance` field and save the model.

        Then, return the :py:attr:`~balance`.

        :param commit: Needs to be True for this method to actually save
                       the model.
        """
        new_balance = self.blockchain.provider.get_balance(self.address)
        if new_balance != self.balance:
            self.balance = new_balance
            if commit:
                self.save()
        return self.balance

    def set_secret_key(self, value):
        """
        Set the encrypted secret key.

        :param value: The secret key to set.
        """
        self.secret_key = binascii.b2a_base64(value).decode()

    def get_secret_key(self):
        """
        Returns the account binary secret key.
        """
        return binascii.a2b_base64(self.secret_key)


@receiver(signals.pre_save, sender=Account)
def account_setup(sender, instance, **kwargs):
    """
    Initial account setup, pre_save signal.

    If the :py:class:`Account` is created with a
    :py:attr:`~Account.secret_key` and without any :py:attr:`~Account.address`
    and the :py:attr:`Blockchain.provider` supports it, then it will fetch the
    :py:attr:`~Account.address` automatically.

    If the :py:class:`Account` is created without an
    :py:attr:`~Account.address` nor a :py:attr:`~Account.secret_key` then it is
    considered that you are creating a new wallet and the
    :py:meth:`djwebdapp.provider.Provider.generate_secret_key()` method will
    generate a py:attr:`~Account.secret_key`.

    If the :py:attr:`~Account.balance` is not set at all, then it will be set
    with the :py:meth:`~Account.refresh_balance()` method.
    """
    if instance.secret_key and not instance.address:
        if hasattr(instance.provider, 'get_address'):
            # for providers which support it
            instance.address = instance.provider.get_address()

    elif not instance.address and not instance.secret_key:
        instance.address, secret_key = \
            instance.blockchain.provider.generate_secret_key()
        instance.set_secret_key(secret_key)

    if not instance.balance:
        instance.refresh_balance(commit=False)


class Blockchain(models.Model):
    """
    Class to represent a blockchain.

    .. py:attribute:: name

        Name of the blockchain.

    .. py:attribute:: provider_class

        Provider class to use for this blockchain.

    .. py:attribute:: is_active

        Boolean to indicate wether this blockchain is active or not.

    .. py:attribute:: description

        Optionnal description for this blockchain.

    .. py:attribute:: chain_id

        Chain ID used to distinguish different EVM blockchains.

    .. py:attribute:: unit

        Unit name of the blockchain, ie. eth, xtz...

    .. py:attribute:: unit_micro

        Unit name of the smallest unit of the blockchain, ie. wei, mutez...

    .. py:attribute:: index_level

        Highest indexed level.

    .. py:attribute:: min_level

        Lowest indexed level

    .. py:attribute:: min_confirmations

        Number of confirmation blocks from confirm to done states.

    .. py:attribute::configurations

        JSON field to store blockchain specific configurations.

    """
    name = models.CharField(
        max_length=100,
    )
    provider_class = models.CharField(
        max_length=255,
        choices=SETTINGS['PROVIDERS'],
    )
    is_active = models.BooleanField(
        default=True,
        help_text='Check to activate this blockchain',
    )
    description = models.TextField(
        blank=True,
        help_text='Free text to describe the blockchain to users',
    )
    chain_id = models.IntegerField(
        blank=True,
        null=True,
    )
    unit = models.CharField(
        max_length=15,
        help_text='Unit name, ie. btc, xtz...',
    )
    unit_micro = models.CharField(
        max_length=15,
        help_text='Unit name, ie. satoshi, mutez...',
    )
    index_level = models.PositiveIntegerField(
        default=None,
        blank=True,
        null=True,
        help_text='Highest indexed level',
    )
    min_level = models.PositiveIntegerField(
        default=None,
        blank=True,
        null=True,
        help_text='Lowest indexed level',
    )
    min_confirmations = models.PositiveIntegerField(
        default=2,
        help_text='Number of confirmation blocks from confirm to done states',
    )
    configuration = models.JSONField(
        default=dict,
        blank=True,
    )

    def __str__(self):
        return self.name

    @property
    def provider_cls(self):
        """ Return the imported provider class. """
        parts = self.provider_class.split('.')
        mod = importlib.import_module('.'.join(parts[:-1]))
        return getattr(mod, parts[-1])

    @property
    def provider(self):
        """ Return a fresh instance of the provider class bound to self. """
        return self.provider_cls(blockchain=self)

    def wait(self):
        """
        Wait for all transactions to be confirmed by min_confirmations blocks.

        For use in between spool and index calls:

        .. code-block:: python

            blockchain.provider.spool()
            blockchain.wait()
            blockchain.provider.index()
        """
        index_level = self.transaction_set.filter(
            state='confirm',
        ).aggregate(
            index_level=Max('level')
        )['index_level']
        if not index_level:
            return  # no transaction to wait.
        self.wait_level(index_level + self.min_confirmations)

    def wait_level(self, level):
        """ Wait for the blockchain head to reach a given level. """
        while self.provider.head < level:
            time.sleep(.1)

    def wait_blocks(self, blocks=None):
        """ Wait for the blockchain head to advance a number of blocks. """
        blocks = blocks or self.min_confirmations
        self.wait_level(self.provider.head + blocks)


class Node(models.Model):
    """
    Blockchain node that we can use to query.

    .. py:attribute:: blockchain

        Foreign key to the :py:class:`~Blockchain` model this node is
        related to.

    .. py:attribute:: name

        Node name, generated from endpoint if empty.

    .. py:attribute:: endpoint

        Node endpoint to query.

    .. py:attribute:: is_active

        Boolean to indicate wether this node is active or not.

    .. py:attribute:: priority

        Nodes with the highest priority will be used first.

    """
    blockchain = models.ForeignKey(
        'Blockchain',
        on_delete=models.CASCADE,
    )
    name = models.CharField(
        max_length=100,
        help_text='Node name, generated from endpoint if empty',
    )
    endpoint = models.URLField(
        help_text='Node endpoint to query',
    )
    is_active = models.BooleanField(
        default=True,
        help_text='Check to activate this node',
    )
    priority = models.IntegerField(
        default=0,
        help_text='Nodes with the highest priority will be used first',
    )


class Explorer(models.Model):
    """
    Blockchain explorer we can use to generate external links.

    .. py:attribute:: blockchain

        Foreign key to the :py:class:`~Blockchain` model this explorer is
        related to.

    .. py:attribute:: name

        Explorer name, generated from url template if empty.

    .. py:attribute:: url_template

        URL template to generate the explorer link.

    .. py:attribute:: is_active

        Boolean to indicate wether this explorer is active or not.

    """
    blockchain = models.ForeignKey(
        'Blockchain',
        on_delete=models.CASCADE,
    )
    name = models.CharField(
        max_length=100,
        help_text='Explorer name, generated from url template if empty',
    )
    url_template = models.CharField(
        max_length=255,
        null=True,
        blank=True,
    )
    is_active = models.BooleanField(
        default=True,
        help_text='Check to activate this explorer',
    )


class TransactionManager(InheritanceManager):
    """
    Manager for the :py:class:`Transaction` model.

    This Model Manager overrides `get_or_create()` and `update_or_create()` to
    handle parent foreign keys dynamically.
    """
    @property
    def parent_fk_column(self):
        """ Return the name of the foreign key to the parent. """
        return self.model._meta.get_ancestor_link(
            self.model._meta.get_parent_list()[0]
        ).name

    def find_or_create(self, lookup_attributes, defaults=None):
        """ Find or create a transaction model. """
        instance = self.filter(**lookup_attributes).first()
        if not instance:
            instance = self.model(
                **lookup_attributes,
                **(defaults or {}),
            )
            instance.save_base(raw=True, force_insert=True)
            instance.refresh_from_db()
            return instance, True
        return instance, False

    def update_or_create(self, *args, **kwargs):
        """ Same as Django's, with dynamic parent fk handled. """
        if (
            self.parent_fk_column not in kwargs
            and f'{self.parent_fk_column}_id' not in kwargs
        ):
            return super().update_or_create(*args, **kwargs)

        defaults = kwargs["defaults"]
        del kwargs["defaults"]
        lookup_attributes = kwargs

        instance, created = self.find_or_create(lookup_attributes, defaults)
        if created:
            return instance, created

        for key, value in defaults.items():
            setattr(instance, key, value)
        instance.save()
        return instance, False

    def get_or_create(self, *args, **kwargs):
        """ Same as Django's, with dynamic parent fk handled. """
        if self.parent_fk_column not in kwargs:
            return super().get_or_create(*args, **kwargs)

        if instance := self.find_or_create(kwargs):
            return instance

        instance = self.filter(**kwargs).first()
        return instance, False


class Transaction(models.Model):
    """
    Transaction superclass, base for all blockchain-specific classes.

    .. py:attribute:: normalizer_class

        Name of the :py:class:`~djwebdapp.normalizers.Normalizer` subclass to
        call to normalize blockchain transactions for contracts subclassing
        this model.

    .. py:attribute:: id

        UUID primary key.

    .. py:attribute:: created_at

        Datetime when the transaction was created.

    .. py:attribute:: updated_at

        Datetime when the transaction was last updated.

    .. py:attribute:: name

        Optional free label name.

    .. py:attribute:: description

        Optional description text.

    .. py:attribute:: blockchain

        Foreign key to the :py:class:`~Blockchain` model this transaction is
        related to.

    .. py:attribute:: level

        Level of the blockchain when the transaction was send.

    .. py:attribute:: hash

        Transaction hash.

    .. py:attribute:: counter

        Transaction counter. Used to avoid replay attack.
        This counter is filled by the blockchain provider.

    .. py:attribute:: nonce

        Transaction nonce. Used to order transactions in the same block.

    .. py:attribute:: gasprice

        Gas price used to send the transaction.

    .. py:attribute:: gas

        Number of gas used to send the transaction.

    .. py:attribute:: last_fail

        If the transaction failed, this field contains the datetime of the
        last failure.

    .. py:attribute:: max_fails

        Number of failures to retry before aborting transaction.

    .. py:attribute:: has_code

        If the transaction is a contract creation, and the transaction
        contains the deployed contract code, this field is set to True.

    .. py:attribute:: metadata

        This field contains the metadata of the transaction.

    .. py:attribute:: normalized

        Enabled when transaction is normalized.
        Set True by the :py:meth:`~Transaction.normalize` method.

    .. py:attribute:: state

        Status of the transaction. By default, the transaction is in the
        `deploy` state. When the transaction is indexed, the state is set to
        `done`.
        The state is updated by the :py:meth:`~state_set` method.

    .. py:attribute:: error

        If the transaction failed, this field contains the error message.

    .. py:attribute:: history

        History of the state of the transaction associated with the datetime
        of the state change.

    .. py:attribute:: amount

        Amount send by the transaction if there is.

    .. py:attribute:: args

        Arguments, appliable to deployments or method calls.

    .. py:attribute:: address

        Contract address, appliable to method calls.

    .. py:attribute:: function

        Function name, if this is a method call.

    .. py:attribute:: sender

        Sender :py:class:`~Account` that send the transaction.

    .. py:attribute:: receiver

        Receiver :py:class:`~Account` , if this is a transfer.

    .. py:attribute:: kind

        Kind of transaction. Can be `contract`, `function` or `transfer`.

    .. py:attribute:: index

        Boolean to indicate wether the indexer should index all
        transactions or not.

    .. py:attribute:: contract

        Smart contract related to this transaction, appliable to method call
        transactions.

    .. py:attribute:: caller

        Transaction that called this one, if any.

    .. py:attribute:: target_contract

        In your own subclasses for function calls, this should be an FK to the
        contract subclass, then :py:meth:`save()` will provision
        :py:attr:`contract` from the parent model of the target contract.

    .. py:attribute:: entrypoint

        In your own subclasses for function calls, this should be the name of
        the function that corresponds to the subclass, for example,
        FA12TezosMint.entrypoint is set to the `"mint"` string.
        Then, :py:meth:`save()` can provision :py:attr:`function` from that.

    .. py:attribute:: contract_name

        In your own subclasses for contract transactions, this should be the
        name of the file that contains smart contract code, without extension.
        The file must be in the sub-directory `contracts` of the django app of
        the model.
        """
    normalizer_class = None
    entrypoint = None
    target_contract = None
    contract_name = None

    id = models.UUIDField(
        primary_key=True,
        editable=False,
        default=uuid.uuid4
    )
    created_at = models.DateTimeField(
        null=True,
        blank=True,
        auto_now_add=True,
        db_index=True,
    )
    updated_at = models.DateTimeField(
        null=True,
        blank=True,
        auto_now=True,
    )
    name = models.CharField(
        help_text='Free label',
        max_length=100,
        null=True,
        blank=True,
    )
    description = models.TextField(
        help_text='Free description text',
        null=True,
        blank=True,
    )
    blockchain = models.ForeignKey(
        'Blockchain',
        on_delete=models.CASCADE,
    )
    level = models.PositiveIntegerField(
        db_index=True,
        null=True,
        blank=True,
    )
    hash = models.CharField(
        max_length=255,
        null=True,
        blank=True,
        db_index=True,
    )
    counter = models.PositiveIntegerField(
        null=True,
    )
    nonce = models.IntegerField(
        default=-1,
        db_index=True,
    )
    number = models.IntegerField(
        default=None,
        blank=True,
        null=True,
    )
    gasprice = models.BigIntegerField(
        blank=True,
        null=True,
    )
    gas = models.BigIntegerField(
        blank=True,
        null=True,
    )
    last_fail = models.DateTimeField(
        null=True,
        blank=True,
    )
    max_fails = models.PositiveIntegerField(
        default=10,
        help_text='Number of failures to retry before aborting transaction',
    )
    has_code = models.BooleanField(
        default=False,
        help_text='Checked if this transaction has smart contract code.',
    )
    metadata = models.JSONField(
        default=dict,
        blank=True,
    )

    STATE_CHOICES = (
        ('held', _('Held')),
        ('deleted', _('Deleted during reorg')),
        ('aborted', _('Aborted')),
        ('deploy', _('To deploy')),
        ('deploying', _('Deploying')),
        ('retry', _('To retry')),
        ('retrying', _('Retrying')),
        ('confirm', _('Deployed to confirm')),
        ('done', _('Confirmed finished')),
    )
    normalized = models.BooleanField(
        default=False,
        help_text='Enabled when transaction is normalized',
        db_index=True,
    )
    state = models.CharField(
        choices=STATE_CHOICES,
        default='deploy',
        max_length=200,
        db_index=True,
    )
    error = models.TextField(blank=True, null=True)
    history = models.JSONField(default=list)
    states = [i[0] for i in STATE_CHOICES]
    amount = models.DecimalField(
        max_digits=80,
        db_index=True,
        decimal_places=18,
        blank=True,
        default=0,
    )
    args = PickledObjectField(
        default=None,
        null=True,
        blank=True,
        help_text='Arguments, appliable to deployments or method calls',
    )
    address = models.CharField(
        max_length=255,
        blank=True,
        null=True,
        db_index=True,
        help_text='Contract address, appliable to method calls',
    )
    # This relation is actually in blockchain specific subclasses.
    # contract = models.ForeignKey(
    #     'self',
    #     on_delete=models.CASCADE,
    #     related_name='call_set',
    #     null=True,
    #     blank=True,
    #     help_text='Smart contract, appliable to method call',
    # )
    function = models.CharField(
        max_length=100,
        db_index=True,
        null=True,
        blank=True,
        help_text='Function name, if this is a method call',
    )
    sender = models.ForeignKey(
        'Account',
        related_name='%(model_name)s_sent',
        null=True,
        blank=True,
        on_delete=models.CASCADE,
    )
    receiver = models.ForeignKey(
        'Account',
        related_name='%(model_name)s_received',
        blank=True,
        null=True,
        on_delete=models.CASCADE,
        help_text='Receiver address, if this is a transfer',
    )
    kind = models.CharField(
        db_index=True,
        blank=True,
        null=True,
        max_length=8,
        choices=(
            ('contract', 'Contract'),
            ('function', 'Function call'),
            ('transfer', 'Transfer'),
        )
    )
    index = models.BooleanField(
        default=True,
        help_text='Wether the indexer should index all transactions or not',
    )
    objects = TransactionManager()

    class Meta:
        unique_together = (
            'blockchain',
            # 'level', ??
            'hash',
            'counter',
            'nonce',
        )
        ordering = ['-created_at']

    def __str__(self):
        if self.name:
            return str(self.name)
        elif self.kind == 'contract' and self.address:
            return self.address
        elif self.hash:
            return str(self.hash)
        elif self.function:
            return f'{self.function}()'
        elif self.amount:
            if getattr(self, 'unit_smallest', None):
                return f'{self.amount}{self.unit_smallest}'
            else:
                return f'{self.amount}'
        else:
            return str(self.pk)

    def state_set(self, state):
        """
        Set :py:attr:`~state` attribute and save it to the :py:attr:`~history`.
        """
        if state == 'done':
            confirmed_level = self.level + self.blockchain.min_confirmations
            head = self.blockchain.provider.head
            if confirmed_level > head:
                self.blockchain.provider.logger.info(
                    f'Set {self} to confirm instead of done'
                )
                state = 'confirm'

        self.state = state
        self.history.append([
            self.state,
            int(datetime.datetime.now().strftime('%s')),
        ])
        self.blockchain.provider.logger.info(
            f'{self}.state={state}'
        )
        self.save()

    @property
    def provider(self):
        """
        Return the blockchain's :py:class:`~djwebdapp.provider.Provider`.
        """
        return self.blockchain.provider

    def deploy(self):
        """
        Deploy the transaction.

        Wrapper around the :py:meth:`djwebdapp.provider.Provider.deploy()`
        method to deploy this trasnsaction.

        In case of exception, it will log the exception, set
        :py:attr:`last_fail`, and count the number of deploy attempts in
        :py:attr:`history`.

        If the deploy attempts count is above the :py:attr:`max_fails` field,
        then set :py:attr:`state` to `aborted`, otherwise, set it to `retry`.
        """
        self.provider.logger.info(f'Deploying {self}')
        self.state_set('deploying')
        try:
            self.sender.provider.deploy(self)
        except Exception:
            self.sender.provider.logger.exception('Deploy fail')
            self.last_fail = timezone.now()

            deploys_since_last_start = 0
            for logentry in reversed(self.history):
                if logentry[0] == 'deploying':
                    deploys_since_last_start += 1
                elif logentry[0] == 'aborted':
                    break

            if deploys_since_last_start >= self.max_fails:
                message = f'Aborting because >= {self.max_fails} failures,'
                self.error = ' '.join([
                    message,
                    'last error:',
                    self.error or '',
                ])
                self.state_set('aborted')
            else:
                self.state_set('retry')
            raise
        else:
            self.last_fail = None
            self.error = ''
            self.state_set('done')
            # indexer is supposed to place it in done

    def save(self, *args, **kwargs):
        """
        Save the transaction in the database.

        Sets :py:attr:`kind` based on the attributes of this transaction and
        :py:attr:`blockchain` from the :py:attr:`sender` automatically.

        Also, provision :py:attr:`~function` and :py:attr:`~contract` from
        :py:attr:`~entrypoint` and :py:attr:`~target_contract`.
        """
        # those would be defined by subclasses
        self_contract = getattr(self, 'contract', None)
        self_target_contract = getattr(self, 'target_contract', None)
        if not self_contract and self_target_contract:
            self.contract = self.target_contract
        if not self.function and self.entrypoint:
            self.function = self.entrypoint

        if not self.kind:
            if not self.function and not self.receiver_id:
                self.kind = 'contract'
            elif self.function:
                self.kind = 'function'
            elif self.amount:
                self.kind = 'transfer'

        if not self.blockchain_id and self.sender_id:
            self.blockchain_id = self.sender.blockchain_id

        if not self.blockchain_id and self.contract_id:
            self.blockchain_id = self.contract.blockchain_id

        return super().save(*args, **kwargs)

    def get_args(self):
        """
        Return the arguments of the transaction.
        """
        return self.args

    def dependency_graph(self):
        """
        Return the dependency graph this transaction is part of, if any.
        """
        dependency = Dependency.objects.filter(
            Q(dependent=self) | Q(dependency=self)
        ).first()
        if dependency:
            return dependency.graph

    def dependency_add(self, transaction):
        """
        Add a transaction that must be deployed before this one.

        :param transaction: The transaction that must be deployed before this
                            one.
        """
        graph = self.dependency_graph() or self
        dependency, _ = Dependency.objects.get_or_create(
            dependency=transaction,
            dependent=self,
            graph=graph,
        )
        return dependency

    def dependency_get(self):
        """
        Return the transaction that must be deployed before this one.
        """
        dependencies = Dependency.objects.filter(
            graph=self.dependency_graph(),
        ).exclude(
            dependent__state='done',
        ).select_related('dependency')
        graph = networkx.DiGraph()
        for dependency in dependencies:
            if dependency.dependency.state == "done":
                graph.add_node(dependency.dependent_id)
            else:
                graph.add_edge(
                    dependency.dependency_id,
                    dependency.dependent_id,
                )

        topological_sort = [node for node in networkx.topological_sort(graph)]
        if len(topological_sort):
            tx_id = topological_sort[0]
            tx = Transaction.objects.filter(
                id=tx_id,
            ).select_subclasses().first()
            return tx

    def normalizer_get(self):
        """
        Return the normalizer class for this transaction.
        """
        if isinstance(self.normalizer_class, str):
            from .normalizers import Normalizer
            return Normalizer._registry[self.normalizer_class]
        elif self.normalizer_class:
            return self.normalizer_class

    def normalize(self):
        """
        Method invoked when normalizing a transaction.

        By default, this relies on
        :py:attr:`~djwebdapp.models.Transaction.normalizer_class`
        """
        contract = self.contract_subclass()
        if not contract:
            return

        normalizer = contract.normalizer_get()
        if not normalizer:
            return

        try:
            normalizer.normalize(self, contract)
        except Exception:
            self.sender.provider.logger.exception('Exception in normalization')
            self.error = traceback.format_exc()
            self.last_fail = timezone.now()
        else:
            self.normalized = True
            self.error = ''
            self.last_fail = None
        self.save()

    @property
    def contract_path(self):
        if not self.contract_name:
            raise Exception('Please contract_name')

        return os.path.join(
            self._meta.app_config.path,
            'contracts',
            self.contract_name,
        )

    def contract_subclass(self):
        """
        Return the subclass of the `.contract` relation.
        """
        if self.kind == 'contract':
            return Transaction.objects.get_subclass(pk=self.pk)

        if not self.contract_id:
            return

        try:
            return Transaction.objects.get_subclass(pk=self.contract_id)
        except Transaction.DoesNotExist:
            pass


class Event(models.Model):
    name = models.CharField(max_length=500)
    args = PickledObjectField(
        default=dict,
        blank=True,
    )
    event_index = models.PositiveBigIntegerField(db_index=True)
    transaction = models.ForeignKey(
        Transaction,
        on_delete=models.CASCADE,
        null=True,
        related_name="transactionevent_set",
    )
    normalized = models.BooleanField(
        default=False,
        db_index=True,
    )

    objects = InheritanceManager()

    def contract_subclass(self):
        return self.transaction.contract_subclass()

    def normalize(self):
        contract = self.contract_subclass()

        if not contract:
            return

        normalizer = contract.normalizer_get()
        if not normalizer:
            return

        try:
            normalizer.normalize_event(self, contract)
        except Exception:
            contract.provider.logger.exception('Exception in event normalizer')
        else:
            self.normalized = True
            self.save()


@receiver(signals.post_save)
def dependency_graph(sender, instance, **kwargs):
    """
    If the transaction is a contract call, then add the contract as dependency.
    """
    if isinstance(instance, Transaction):
        # contract_id are defined in blockchain specific subclasses
        # getattr here prevents AttributeError: 'Transaction' object has no
        # attribute 'contract_id'
        if instance.function and getattr(instance, 'contract_id', None):
            instance.dependency_add(instance.contract)


class Dependency(models.Model):
    """
    A dependency between two transactions.

    When a transaction is deployed, it is checked if it has a dependency.
    If it does, the transaction is not deployed, but the dependency is
    checked instead. If the dependency is deployed, the transaction
    is deployed. Otherwise, the transaction is not.

    .. py:attribute:: dependent

        The :py:class:`~djwebdapp.models.Transaction` that depends on the
        :py:attr:`~dependency` one.

    .. py:attribute:: dependency

        The :py:class:`~djwebdapp.models.Transaction` that is required to be
        deployed before the :py:attr:`~dependent` one.

    .. py:attribute:: graph

        The :py:class:`~djwebdapp.models.Transaction` that this graph
        was created for. This is used for performance reasons, to not
        have to load the full dependency table for every transaction.

    .. py:attribute:: created_at

        The time the dependency was created.

    .. py:attribute:: updated_at

        The time the dependency was updated.

    """
    dependent = models.ForeignKey(
        'Transaction',
        on_delete=models.CASCADE,
    )
    dependency = models.ForeignKey(
        'Transaction',
        on_delete=models.CASCADE,
        related_name='dependent_set',
    )
    # This serves purely for performance, to not have to load the full
    # dependency table for every transaction.
    graph = models.ForeignKey(
        'Transaction',
        on_delete=models.CASCADE,
        related_name='graph',
        help_text='The transaction this graph was created for',
    )
    created_at = models.DateTimeField(
        null=True,
        blank=True,
        auto_now_add=True,
    )
    updated_at = models.DateTimeField(
        null=True,
        blank=True,
        auto_now=True,
    )



# Generated by Django 4.1.1 on 2022-09-11 16:30

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("djwebdapp", "0007_alter_transaction_amount"),
    ]

    operations = [
        migrations.AddField(
            model_name="transaction",
            name="index",
            field=models.BooleanField(
                default=True,
                help_text="Wether the indexer should index all transactions or not",
            ),
        ),
    ]



Contract = client.eth.contract(
    abi=contract.abi,
    address=contract.address,
)

tx = Contract.functions.mint(client.eth.default_account, 300).transact()
receipt = client.eth.wait_for_transaction_receipt(tx)
txhash = receipt['transactionHash'].hex()

blockchain.provider.index()
indexed = blockchain.provider.transaction_class.objects.get(hash=txhash)
assert indexed.args, 2

blockchain.provider.normalize()

normalized = contract.fa12ethereummint_set.get(hash=txhash)
assert normalized.mint_amount == 300



# Generated by Django 4.1.1 on 2023-02-03 14:39

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ("djwebdapp_tezos", "0005_tezoscall_tezoscontract"),
        ("djwebdapp_multisig", "0001_initial"),
        ("djwebdapp", "0013_transaction_normalized_dependency"),
    ]

    operations = [
        migrations.CreateModel(
            name="Fa2Contract",
            fields=[
                (
                    "tezostransaction_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to="djwebdapp_tezos.tezostransaction",
                    ),
                ),
                ("metadata_uri", models.CharField(max_length=500)),
                (
                    "manager",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="fa2_administered",
                        to="djwebdapp.account",
                    ),
                ),
                (
                    "multisig",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        to="djwebdapp_multisig.multisigcontract",
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
            bases=("djwebdapp_tezos.tezoscontract",),
        ),
        migrations.CreateModel(
            name="UpdateProxyCall",
            fields=[
                (
                    "tezostransaction_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to="djwebdapp_tezos.tezostransaction",
                    ),
                ),
                ("remove_proxy", models.BooleanField(default=False)),
                (
                    "proxy",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="fa2contract_proxy_set",
                        to="djwebdapp.account",
                    ),
                ),
            ],
            bases=("djwebdapp_tezos.tezoscall",),
        ),
        migrations.CreateModel(
            name="MintCall",
            fields=[
                (
                    "tezostransaction_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to="djwebdapp_tezos.tezostransaction",
                    ),
                ),
                ("token_id", models.PositiveBigIntegerField()),
                ("token_amount", models.PositiveBigIntegerField()),
                ("metadata_uri", models.CharField(max_length=500)),
                (
                    "owner",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="mintcall_owner_set",
                        to="djwebdapp.account",
                    ),
                ),
            ],
            bases=("djwebdapp_tezos.tezoscall",),
        ),
        migrations.CreateModel(
            name="Fa2Token",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("token_id", models.IntegerField()),
                ("metadata_uri", models.CharField(max_length=500)),
                (
                    "contract",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="djwebdapp_fa2.fa2contract",
                    ),
                ),
            ],
            options={
                "unique_together": {("token_id", "contract")},
            },
        ),
        migrations.CreateModel(
            name="Balance",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("amount", models.PositiveBigIntegerField(default=0)),
                (
                    "account",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="fa2_balances",
                        to="djwebdapp.account",
                    ),
                ),
                (
                    "token",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="djwebdapp_fa2.fa2token",
                    ),
                ),
            ],
            options={
                "unique_together": {("token", "account")},
            },
        ),
    ]



from django.apps import AppConfig
from django.utils.module_loading import autodiscover_modules


class DjwebdappConfig(AppConfig):
    name = 'djwebdapp'

    def ready(self):
        autodiscover_modules('normalizers')



from dataclasses import dataclass
from pymich.michelson_types import Tezos, BaseContract, Contract, Nat, Mutez


@dataclass(kw_only=True)
class Callee(BaseContract):
    counter: Nat
    price: Mutez

    def set_counter(self, new_counter: Nat):
        if Tezos.amount != self.price:
            raise Exception("Not the right price")

        self.counter = new_counter
        set_counter_callback_entrypoint = Contract[Nat](Tezos.sender, "%set_counter_callback")
        self.ops = self.ops.push(
            set_counter_callback_entrypoint,
            Tezos.amount,
            new_counter,
        )

    def get_counter(self) -> Nat:
        return self.counter



# let's mint some sweet tokens
client_mint = client.contract(contract.address).mint(
    'tz1KqTpEZ7Yob7QbPE4Hy4Wo8fHG8LhKxZSx',
    300,
).send(min_confirmations=2)

# should this really be needed?
blockchain.wait_blocks()

blockchain.provider.index()
indexed = blockchain.provider.transaction_class.objects.get(hash=client_mint.hash())
assert indexed.args, dict(
    _to='tz1KqTpEZ7Yob7QbPE4Hy4Wo8fHG8LhKxZSx',
    value=300,
)

blockchain.provider.normalize()

normalized = contract.fa12tezosmint_set.get(hash=client_mint.hash())
assert normalized.mint_amount == 300

# normalized transaction is a subclass of indexed transaction
assert normalized.tezostransaction_ptr == indexed



from datetime import datetime
from unittest.mock import Mock

import pytest

from django.contrib import auth
from djwebdapp_rest_framework import views
from rest_framework.test import APIClient
from pytezos.crypto.key import Key


@pytest.mark.django_db
def test_airgap_wallet():
    data = {
        'message': '1644848622',
        'signed_message': 'edsigts6uAiJAU5oRvbBxBFEKMTk21ozxX7ngjXfSinpdhQHKKgVNgd37aRaF6tcvbFUrffzLRYWdSfmt7XrJ638DHZgECxQkee',  # noqa
        'public_key': '3279a2c27e8bc67f43b76eb7f8d57cd47591edeede0387e8164d55706af5c8da'
    }
    url = '/api-token-auth/'

    response = APIClient().post(url, data, format='json')
    assert response.data == {'error': 'Message expired'}


@pytest.mark.django_db
def test_temple_wallet():
    data = {
        'message': '1645436647',
        'signed_message': 'edsigtedLyAiny2uNz5VQpoWn2PdFm8aoLURRR9RByMZGyjGRbGQ63xrFBBpzhvdQZodU4sy3rNTodqWGDqTUDiFf2k9Hxe2ynx',  # noqa
        'public_key': 'edpkuhJRth9s12SApCJBwgdCpcHbjxqth6HXCZBwAnheSoysqKT23R'
    }
    url = '/api-token-auth/'

    response = APIClient().post(url, data, format='json')
    assert response.data == {'error': 'Message expired'}, response.data


@pytest.mark.django_db
def test_kukai_wallet():
    data = {
        'message': '1645439434',
        'signed_message': 'spsig1SbxrnSbkn76G98GZk5BUkYL71dLJuK7E21VjdXtfQn93ad2f8y7Uv5TCXXihf9ybtHTih9PDHetTn8jQ6dcRebJv9i6AZ',  # noqa
        'public_key': 'sppk7d5Q7pPf4u48EyrtHSCCuXx2tBsVRbWXpxGbn8UbUgEQyqBTxhQ'
    }
    url = '/api-token-auth/'

    response = APIClient().post(url, data, format='json')
    assert response.data == {'error': 'Message expired'}


@pytest.mark.django_db
def test_invalid_timestamp():
    data = {
        'message': 'abc',
        'signed_message': 'spsig1SbxrnSbkn76G98GZk5BUkYL71dLJuK7E21VjdXtfQn93ad2f8y7Uv5TCXXihf9ybtHTih9PDHetTn8jQ6dcRebJv9i6AZ',  # noqa
        'public_key': 'sppk7d5Q7pPf4u48EyrtHSCCuXx2tBsVRbWXpxGbn8UbUgEQyqBTxhQ'
    }
    url = '/api-token-auth/'

    response = APIClient().post(url, data, format='json')
    assert response.data == {'error': 'Could not parse timestamp "abc"'}




@pytest.mark.django_db
def test_login(monkeypatch):
    message = str(round(datetime.now().timestamp()))
    encoded_message = views.encode_message(message)
    key = Key.from_encoded_key('edsk3gUfUPyBSfrS9CCgmCiQsTCHGkviBDusMxDJstFtojtc1zcpsh')
    signed_message = key.sign(encoded_message)
    key.verify(signed_message, encoded_message)
    data = {
        'message': message,
        'signed_message': signed_message,
        'public_key': key.public_key(),
        'encoded_message': encoded_message,
    }
    url = '/api-token-auth/'

    # we require a user model with public_key and public_key_hash attributes
    # let's mock everything
    User = auth.get_user_model()
    user = User()
    user.public_key_hash = 'tz1KqTpEZ7Yob7QbPE4Hy4Wo8fHG8LhKxZSx'
    user.username = 'tz1KqTpEZ7Yob7QbPE4Hy4Wo8fHG8LhKxZSx'
    User._default_manager.get_or_create = lambda *a, **k: (user, True)
    User._default_manager.make_random_password = lambda: 'test'
    monkeypatch.setattr(views, 'User', User)

    response = APIClient().post(url, data, format='json')

    assert 'token' in response.data
    assert response.data['user_id'] == 1
    assert response.data['public_key_hash'] == key.public_key_hash()
    assert response.data['username'] == key.public_key_hash()



from django.contrib.auth import get_user_model
from django.db import models
from djwebdapp.models import Account
from djwebdapp_tezos.models import TezosCall, TezosContract


User = get_user_model()


class MultisigContract(TezosContract):
    contract_name = "multisig"
    admin = models.ForeignKey(
        Account,
        on_delete=models.CASCADE,
        related_name="multisigs_administered",
    )
    owner = models.ForeignKey(
        User,
        on_delete=models.CASCADE,
        null=True,
    )

    def get_init_storage(self):
        init_storage = self.get_contract_interface().storage.dummy()
        init_storage["admins"] = [self.admin.address]

        return init_storage


class AddAuthorizedContractCall(TezosCall):
    entrypoint = "addAuthorizedContract"
    contract_to_authorize = models.ForeignKey(
        "djwebdapp_tezos.TezosTransaction",
        on_delete=models.CASCADE,
        related_name="add_authorized_contract_calls",
        blank=True,
        null=True,
    )

    def get_args(self):
        return [self.contract_to_authorize.address]



from datetime import datetime, timedelta

from django.contrib import auth
from django.db.utils import IntegrityError

from rest_framework.authentication import BasicAuthentication
from rest_framework.authtoken.models import Token
from rest_framework.authtoken.views import ObtainAuthToken
from rest_framework.response import Response

from pytezos.crypto.encoding import base58_encode
from pytezos.crypto.key import Key


User = auth.get_user_model()


def string_length_in_hexa_over_four_bytes(message: str) -> str:
    """
    This function takes a string as input and returns its length encoded
    in hexadecimal over 4 bytes.

    - input: a message
    - output: the message length represented in hex over 4 bytes

    example:
    - input: `foobar`
    - output:
        len("abcdefghijklmnopqrstuvwxyz") == 26
        hex(len("abcdefghijklmnopqrstuvwxyz")) == "0x1a"
        hex(len("abcdefghijklmnopqrstuvwxyz"))[2:] == "1a"
        hex(len("abcdefghijklmnopqrstuvwxyz"))[2:].zfill(8) == "0000001a"
    """
    return hex(len(message))[2:].zfill(8)


def encode_message(message: str) -> str:
    """
    Encodes a message to check the signature sent by the backend against
    """
    message = f'Tezos Signed Message: {message}'
    length = string_length_in_hexa_over_four_bytes(message)
    return '0x0501' + length + message.encode().hex()


def decode_airgap_public_key(raw_public_key: str) -> str:
    public_key_raw = bytes.fromhex(raw_public_key)
    return base58_encode(public_key_raw, b'edpk').decode('ascii')


class WalletLogin(ObtainAuthToken):
    authentication_classes = [BasicAuthentication]

    def post(self, request, *args, **kwargs):
        self.serializer_class(
            data=request.data,
            context=dict(request=request)
        )

        try:
            public_key_alphanum = decode_airgap_public_key(
                request._data['public_key']
            )
        except ValueError:
            public_key_alphanum = request._data['public_key']

        message = request._data['message']
        signed_message = request._data['signed_message']

        try:
            message_date = datetime.utcfromtimestamp(int(message))
        except ValueError:
            return Response(
                {
                    'error': f'Could not parse timestamp "{message}"',
                }
            )

        current_timestamp = datetime.utcnow()
        allowed_delta = timedelta(minutes=10)

        try:
            public_key = Key.from_encoded_key(public_key_alphanum)
            encoded_message = encode_message(message)
            public_key.verify(signed_message, encoded_message)
        except ValueError:
            return Response(
                {
                    'error': 'Failed to validate message signature',
                }
            )

        if current_timestamp - message_date > allowed_delta:
            return Response(
                {
                    'error': 'Message expired',
                }
            )

        try:
            user, created = User._default_manager.get_or_create(
                public_key_hash=public_key.public_key_hash(),
                public_key=public_key_alphanum,
                username=public_key.public_key_hash(),
            )
        except IntegrityError:
            return Response(
                {
                    'error': 'User creation failed',
                }
            )

        if created:
            user.password = User.objects.make_random_password()
            user.save()

        token, created = Token.objects.get_or_create(user=user)
        return Response(
            {
                'token': token.key,
                'user_id': user.pk,
                'public_key_hash': user.public_key_hash,
                'username': user.username,
            }
        )



# Generated by Django 4.1.1 on 2023-02-03 14:38

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ("djwebdapp", "0012_transaction_number"),
    ]

    operations = [
        migrations.AddField(
            model_name="transaction",
            name="normalized",
            field=models.BooleanField(
                default=False, help_text="Enabled when transaction is normalized"
            ),
        ),
        migrations.CreateModel(
            name="Dependency",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("created_at", models.DateTimeField(auto_now_add=True, null=True)),
                ("updated_at", models.DateTimeField(auto_now=True, null=True)),
                (
                    "dependency",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="dependent_set",
                        to="djwebdapp.transaction",
                    ),
                ),
                (
                    "dependent",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="djwebdapp.transaction",
                    ),
                ),
                (
                    "graph",
                    models.ForeignKey(
                        help_text="The transaction this graph was created for",
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="graph",
                        to="djwebdapp.transaction",
                    ),
                ),
            ],
        ),
    ]



from django.core.management.base import BaseCommand

from djwebdapp.models import Blockchain


class Command(BaseCommand):
    help = '''
    Import a contract from tzkt.

    Example usage:

        ./manage.py history_download 'Tezos Mainnet' <address>
    '''

    def add_arguments(self, parser):
        parser.add_argument(
            'blockchain',
            type=str,
            help='Name of the blockchain to import into'
        )
        parser.add_argument(
            'target',
            type=str,
            help='Id of the contract to import',
        )

    def handle(self, *args, **options):
        blockchain = Blockchain.objects.get(name=options['blockchain'])
        blockchain.provider.download(target=options['target'])



"""
Normalizer classes are optionnal helpers to easy contract indexation.
"""


class Normalizer:
    """
    Base Normalizer class.
    """
    _registry = {}
    deploy_method_name = 'deploy'

    def __init_subclass__(cls, /, **kwargs):
        super().__init_subclass__(**kwargs)
        cls._registry[cls.__name__] = cls

    @classmethod
    def normalize(cls, transaction, contract):
        normalizer = cls()
        if transaction.kind == 'function':
            callback_name = transaction.function
        elif transaction.kind == 'contract':
            callback_name = cls.deploy_method_name

        callback = getattr(normalizer, callback_name, None)
        if callback:
            callback(transaction, contract)

    @classmethod
    def normalize_event(cls, event, contract):
        normalizer = cls()
        callback_name = event.name
        callback = getattr(normalizer, callback_name, None)
        if callback:
            callback(event, contract)



from dataclasses import dataclass
from pymich.michelson_types import Address, BaseContract, Contract, Mutez, Nat

@dataclass(kw_only=True)
class Proxy(BaseContract):
    callee: Address
    admin: Address
    counter: Nat

    def set_counter(self, new_counter: Nat, price: Mutez):
        increment_counter_entrypoint = Contract[Nat](self.callee, "%set_counter")
        self.ops = self.ops.push(
            increment_counter_entrypoint,
            price,
            new_counter,
        )

    def set_counter_callback(self, new_counter: Nat):
        self.counter = new_counter

    def getCallee(self) -> Address:
        return self.callee



# Generated by Django 3.2.18 on 2023-04-08 13:33

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('djwebdapp_ethereum', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='EthereumCall',
            fields=[
            ],
            options={
                'proxy': True,
                'indexes': [],
                'constraints': [],
            },
            bases=('djwebdapp_ethereum.ethereumtransaction',),
        ),
        migrations.CreateModel(
            name='EthereumContract',
            fields=[
            ],
            options={
                'proxy': True,
                'indexes': [],
                'constraints': [],
            },
            bases=('djwebdapp_ethereum.ethereumtransaction',),
        ),
    ]



from djwebdapp_example_ethereum.models import FA12Ethereum, FA12EthereumMint

# create a normalized contract object
contract = FA12Ethereum.objects.create(
    blockchain=blockchain,
    sender=admin,
    token_name='Your Token',
    token_symbol='YT',
)

# make a couple of mint calls
first_mint = FA12EthereumMint.objects.create(
    target_contract=contract,
    sender=admin,
    mint_account=admin,
    mint_amount=100,
)
second_mint = FA12EthereumMint.objects.create(
    target_contract=contract,
    sender=admin,
    mint_account=admin,
    mint_amount=200,
)

# let's deploy each transaction one by one
# run ./manage.py spool in a bash loop instead of doing this in python
import time
while blockchain.provider.spool():
    time.sleep(.1)

contract.refresh_from_db()
print(contract.level, contract.hash, contract.address)

first_mint.refresh_from_db()
print(first_mint.level, first_mint.hash)

second_mint.refresh_from_db()
print(second_mint.level, second_mint.hash)

# Given all these transactions are sent from the same account, each was done in
# a different block to prevent noonce issues
assert contract.level < first_mint.level < second_mint.level



# Generated by Django 4.0.5 on 2022-06-14 16:30

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('djwebdapp', '0004_alter_transaction_options'),
    ]

    operations = [
        migrations.AlterUniqueTogether(
            name='transaction',
            unique_together=set(),
        ),
        migrations.AddField(
            model_name='transaction',
            name='counter',
            field=models.PositiveIntegerField(null=True),
        ),
        migrations.AddField(
            model_name='transaction',
            name='nonce',
            field=models.PositiveIntegerField(null=True),
        ),
        migrations.AlterUniqueTogether(
            name='transaction',
            unique_together={('blockchain', 'hash', 'counter', 'nonce')},
        ),
        migrations.RemoveField(
            model_name='transaction',
            name='txgroup',
        ),
    ]



from djwebdapp.models import Account
from djwebdapp.normalizers import Normalizer

from djwebdapp_example_ethereum.models import (
    FA12Ethereum,
    FA12EthereumMint,
    FA12EthereumBalance,
    FA12EthereumBalanceMovement,
)


class FA12EthereumNormalizer(Normalizer):
    def mint(self, call, contract):
        account, _ = Account.objects.get_or_create(
            address=call.args['account'],
        )
        call, _ = FA12EthereumMint.objects.update_or_create(
            ethereumtransaction_ptr_id=call.id,
            defaults=dict(
                target_contract=contract,
                mint_account=account,
                mint_amount=call.args['amount'],
            )
        )
        balance, _ = FA12EthereumBalance.objects.get_or_create(
            account=account,
            fa12=contract,
        )
        balance.balance += call.mint_amount
        balance.save()

    def Mint(self, event, contract):
        account, _ = Account.objects.get_or_create(
            address=event.args['to_'],
            blockchain=contract.blockchain,
        )
        FA12EthereumBalanceMovement.objects.update_or_create(
            event=event,
            defaults=dict(
                fa12=contract,
                account=account,
                amount=event.args['amount_'],
            )
        )

    def Mint__reorg(self, event, contract):
        balance_movement = FA12EthereumBalanceMovement.objects.get(
            fa12=contract,
            event=event,
        )
        balance_movement.delete()



from rest_framework import generics
from fa2.models import Fa2Token
from fa2.serializers import Fa2TokenSerializer


class Fa2TokenRetrieveAPIView(generics.RetrieveAPIView):
    """
    API endpoint that shows token auctions
    """
    queryset = Fa2Token.objects.all()
    serializer_class = Fa2TokenSerializer

    def get_object(self):
        return (
            Fa2Token.objects.filter(
                contract__origination__address=self.kwargs["token_address"],
                token_id=self.kwargs["token_id"],
            )
            .select_related("contract__origination")
            .select_related("metadata")
            .select_related("metadata__artifact_asset")
            .select_related("metadata__display_asset")
            .select_related("metadata__thumbnail_asset")
            .select_related("metadata__artifact_asset__ipfs_gateway")
            .select_related("metadata__display_asset__ipfs_gateway")
            .select_related("metadata__thumbnail_asset__ipfs_gateway")
            .prefetch_related("metadata__tags")
            .prefetch_related("metadata__creators")
            .prefetch_related("metadata__attributes")
            .prefetch_related("metadata__royalties")
            .prefetch_related("metadata__attributes")
            .first()
        )



# Generated by Django 3.2.18 on 2023-04-08 19:20

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('djwebdapp_ethereum', '0002_ethereumcall_ethereumcontract'),
    ]

    operations = [
        migrations.AddField(
            model_name='ethereumtransaction',
            name='caller',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, related_name='_internal_calls', to='djwebdapp_ethereum.ethereumtransaction'),
        ),
    ]



# Generated by Django 3.2.18 on 2023-04-11 14:08

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('djwebdapp_tezos', '0005_tezoscall_tezoscontract'),
        ('djwebdapp', '0015_alter_transaction_state'),
    ]

    operations = [
        migrations.CreateModel(
            name='FA12Tezos',
            fields=[
                ('tezostransaction_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='djwebdapp_tezos.tezostransaction')),
                ('token_name', models.CharField(max_length=200)),
                ('token_symbol', models.CharField(max_length=10)),
            ],
            bases=('djwebdapp_tezos.tezoscontract',),
        ),
        migrations.CreateModel(
            name='FA12TezosMint',
            fields=[
                ('tezostransaction_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='djwebdapp_tezos.tezostransaction')),
                ('mint_amount', models.PositiveIntegerField()),
                ('mint_account', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='djwebdapp.account')),
                ('target_contract', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='djwebdapp_example_tezos.fa12tezos')),
            ],
            bases=('djwebdapp_tezos.tezoscall',),
        ),
        migrations.CreateModel(
            name='FA12TezosBalance',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('balance', models.PositiveIntegerField(default=0)),
                ('account', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='djwebdapp.account')),
                ('fa12', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='djwebdapp_example_tezos.fa12tezos')),
            ],
            options={
                'unique_together': {('fa12', 'account')},
            },
        ),
    ]



from djwebdapp.models import Blockchain

blockchain, _ = Blockchain.objects.get_or_create(
    name='Ethereum Local with event provider',
    provider_class='djwebdapp_ethereum.provider.EthereumEventProvider',
    min_confirmations=0,
)

blockchain.node_set.get_or_create(
    endpoint='http://ethlocal:8545',
)



from django.db import models

from djwebdapp.models import Event, Transaction
from djwebdapp.normalizers import Normalizer


class EthereumTransaction(Transaction):
    """
    Base model for Ethereum transactions.

    .. py:attribute:: abi

        Smart contract ABI code.

    .. py:attribute:: bytecode

        Smart contract bytecode.
    """
    contract = models.ForeignKey(
        'self',
        on_delete=models.CASCADE,
        related_name='call_set',
        null=True,
        blank=True,
        help_text='Smart contract, appliable to method call',
    )
    caller = models.ForeignKey(
        'self',
        on_delete=models.CASCADE,
        null=True,
        related_name='_internal_calls',
    )
    abi = models.JSONField(
        default=dict,
        blank=True,
        null=True,
        help_text='Smart contract ABI, if this is a smart contract',
    )
    input = models.TextField(
        blank=True,
        null=True,
        help_text='Input hex string if any',
    )
    bytecode = models.TextField(
        blank=True,
        null=True,
        help_text='Contract bytecode if this is a smart contract to deploy',
    )

    def save(self, *args, **kwargs):
        """
        Sets :py:attr:`~djwebdapp.models.Transaction.has_code` if
        :py:attr:`bytecode` is set.
        """
        if self.bytecode and self.abi:
            self.has_code = True
        return super().save(*args, **kwargs)

    @property
    def receipt(self):
        return self.provider.client.eth.get_transaction_receipt(self.hash)

    @property
    def contract_ci(self):
        return self.provider.client.eth.contract(
            address=self.contract.address,
            abi=self.contract.abi,
        )

    def get_event(self, event_name):
        event = getattr(
            self.contract_ci.events,
            event_name,
        )()
        return event.process_receipt(self.receipt)


class EthereumContract(EthereumTransaction):
    """
    Base model class for Ethereum Contracts.

    .. py:attribute:: contract_name

        Name of the contract files, they are expected to be found in the
        ``contracts`` sub-directory of the Django App that holds the model that
        is inheriting from this class (your app)
    """

    contract_name = None
    normalizer_class = Normalizer

    class Meta:
        proxy = True

    def save(self, *args, **kwargs):
        """
        Sets :py:attr:`~djwebdapp_ethereum.models.EthereumTransaction.abi` and
        :py:attr:`~djwebdapp_ethereum.models.EthereumTransaction.bytecode` if
        :py:attr:`contract_name` is set.
        """
        if self.contract_name and not self.abi:
            with open(self.contract_path + '.abi', 'r') as f:
                self.abi = f.read()

        if self.contract_name and not self.bytecode:
            with open(self.contract_path + '.bin', 'r') as f:
                self.bytecode = f.read()

        return super().save(*args, **kwargs)

    @property
    def is_internal(self):
        return bool(self.caller_id)

    @property
    def internal_calls(self):
        if self.is_internal:
            txgroup_internal_calls_qs = self.caller._internal_calls
        else:
            txgroup_internal_calls_qs = self._internal_calls

        tx_internal_calls_qs = txgroup_internal_calls_qs.filter(
            nonce__gte=self.nonce if self.nonce else 0,
            sender__address=self.contract.address,
        )
        return tx_internal_calls_qs.order_by('nonce').all()


class EthereumCall(EthereumTransaction):
    """
    Base model class for Ethereum contract function calls.
    """

    class Meta:
        proxy = True


class EthereumEvent(Event):
    contract = models.ForeignKey(
        EthereumContract,
        on_delete=models.CASCADE,
        related_name="contractevent_set",
    )

    def contract_subclass(self):
        return Transaction.objects.get_subclass(pk=self.contract.pk)



# Generated by Django 4.1.1 on 2022-09-16 14:34

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("djwebdapp", "0010_alter_transaction_nonce"),
    ]

    operations = [
        migrations.RenameField(
            model_name="blockchain",
            old_name="max_level",
            new_name="index_level",
        ),
    ]



import pytest
from djwebdapp.models import Blockchain, Transaction


@pytest.mark.django_db
def test_get_next_deploy():
    """
    tx_1
    /  \
  tx_2 tx_3
    \  /
    tx_4
    """
    blockchain = Blockchain.objects.create(
        provider_class='djwebdapp_tezos.provider.TezosProvider',
    )

    tx1 = Transaction.objects.create(blockchain=blockchain, state="deploy", name='tx1')
    tx2 = Transaction.objects.create(blockchain=blockchain, state="deploy", name='tx2')
    tx3 = Transaction.objects.create(blockchain=blockchain, state="deploy", name='tx3')
    tx4 = Transaction.objects.create(blockchain=blockchain, state="deploy", name='tx4')

    tx1.dependency_add(tx2)
    tx1.dependency_add(tx3)
    tx2.dependency_add(tx4)
    tx3.dependency_add(tx4)

    assert tx1.dependency_get() == tx4
    tx4.state = 'done'
    tx4.save()

    assert tx1.dependency_get() == tx2
    tx2.state = 'done'
    tx2.save()

    assert tx1.dependency_get() == tx3
    tx3.state = 'done'
    tx3.save()

    assert tx1.dependency_get() == tx1



# Generated by Django 4.1.1 on 2023-02-03 14:38

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("djwebdapp_tezos", "0005_tezoscall_tezoscontract"),
        ("djwebdapp", "0013_transaction_normalized_dependency"),
    ]

    operations = [
        migrations.CreateModel(
            name="MultisigContract",
            fields=[
                (
                    "tezostransaction_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to="djwebdapp_tezos.tezostransaction",
                    ),
                ),
                (
                    "admin",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="multisigs_administered",
                        to="djwebdapp.account",
                    ),
                ),
                (
                    "owner",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            bases=("djwebdapp_tezos.tezoscontract",),
        ),
        migrations.CreateModel(
            name="AddAuthorizedContractCall",
            fields=[
                (
                    "tezostransaction_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to="djwebdapp_tezos.tezostransaction",
                    ),
                ),
                (
                    "contract_to_authorize",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="add_authorized_contract_calls",
                        to="djwebdapp_tezos.tezostransaction",
                    ),
                ),
            ],
            bases=("djwebdapp_tezos.tezoscall",),
        ),
    ]



from djwebdapp.models import Account

from djwebdapp_fa2.models import Balance, Fa2Token
from djwebdapp.normalizers import Normalizer


class Fa2Normalizer(Normalizer):
    def mint(self, call, contract):
        token, _ = Fa2Token.objects.update_or_create(
            contract=contract,
            token_id=call.args["token_id"],
        )
        owner_account, _ = Account.objects.get_or_create(
            blockchain=call.blockchain,
            address=call.args["owner"],
        )
        Balance.objects.update_or_create(
            account=owner_account,
            token=token,
            defaults=dict(
                amount=call.args["amount_"],
            ),
        )

    def transfer(self, call, contract):
        from_account, _ = Account.objects.get_or_create(
            blockchain=call.blockchain,
            address=call.args[0]["from_"],
        )

        token = contract.fa2token_set.filter(
            token_id=call.args[0]["txs"][0]["token_id"],
        ).first()

        from_balance, _ = Balance.objects.get_or_create(
            account=from_account,
            token=token,
        )

        for transfer in call.args[0]["txs"]:
            to_account, _ = Account.objects.get_or_create(
                blockchain=call.blockchain,
                address=transfer["to_"],
            )
            to_balance, created = Balance.objects.get_or_create(
                account=to_account,
                token=token,
            )
            if created:
                to_balance.amount = 0

            from_balance.amount -= transfer["amount"]

            to_balance.amount += transfer["amount"]
            to_balance.save()

        from_balance.save()

    def burn(self, call, contract):
        token = contract.fa2token_set.filter(
            token_id=call.args["token_id"],
        ).first()

        burner_balance, _ = Balance.objects.get_or_create(
            account=call.sender,
            token=token,
        )

        burner_balance.amount -= call.args["token_amount"]
        burner_balance.save()



# Generated by Django 4.2.7 on 2023-11-28 14:49

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):
    dependencies = [
        ("djwebdapp", "0017_event"),
        ("djwebdapp_ethereum", "0003_ethereumtransaction_caller"),
    ]

    operations = [
        migrations.CreateModel(
            name="EthereumEvent",
            fields=[
                (
                    "event_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to="djwebdapp.event",
                    ),
                ),
                (
                    "contract",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="contractevent_set",
                        to="djwebdapp_ethereum.ethereumcontract",
                    ),
                ),
            ],
            bases=("djwebdapp.event",),
        ),
    ]



# Generated by Django 5.2 on 2025-11-13 17:40

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("djwebdapp", "0018_blockchain_chain_id"),
    ]

    operations = [
        migrations.AlterField(
            model_name="event",
            name="event_index",
            field=models.PositiveBigIntegerField(db_index=True),
        ),
        migrations.AlterField(
            model_name="event",
            name="normalized",
            field=models.BooleanField(db_index=True, default=False),
        ),
        migrations.AlterField(
            model_name="transaction",
            name="created_at",
            field=models.DateTimeField(auto_now_add=True, db_index=True, null=True),
        ),
        migrations.AlterField(
            model_name="transaction",
            name="hash",
            field=models.CharField(
                blank=True, db_index=True, max_length=255, null=True
            ),
        ),
        migrations.AlterField(
            model_name="transaction",
            name="nonce",
            field=models.IntegerField(db_index=True, default=-1),
        ),
        migrations.AlterField(
            model_name="transaction",
            name="normalized",
            field=models.BooleanField(
                db_index=True,
                default=False,
                help_text="Enabled when transaction is normalized",
            ),
        ),
    ]
